import sys
import os
import numpy as np
import nibabel as nib
import vnmrjpy as vj
import warnings
import matplotlib.pyplot as plt

class KspaceMaker():
    """Class to build the k-space from the raw fid data based on procpar.

    Raw fid_data is numpy.ndarray(blocks, traces * np) format. Should be
    untangled based on 'seqcon' or 'seqfil' parameters.
    seqcon chars refer to (echo, slice, Pe1, Pe2, Pe3)

    Should support compressed sensing
    In case of Compressed sensing the reduced kspace is filled with zeros
    to reach the intended final shape

    Leave rest of reconstruction to other classes/functions

    INPUT:  fid data = np.ndarra([blocks, np*traces])
            fid header
            procpar

    METHODS:
            make(): 
                return kspace = nump.ndarray\
                            ([rcvrs, phase, read, slice, echo*time])

    """
    def __init__(self, fid_data, fidheader, procpar, verbose=False):
        """Reads procpar"""

        def _get_arrayed_AP(p):
            """check for arrayed pars in procpar

            Return: dictionary {par : array_length}
            """

            AP_dict = {}
            for par in ['tr', 'te', 'fa']:
            
                pass

            return AP_dict

        self.p = vj.io.ProcparReader(procpar).read()
        self.fid_header = fidheader
        self.rcvrs = str(self.p['rcvrs']).count('y')
        self.arrayed_AP = _get_arrayed_AP(self.p)
        apptype = self.p['apptype']
        self.config = vj.config
        self.verbose = verbose
        self.procpar = procpar
        # decoding skipint parameter
        # TODO
        # final kspace shape from config file
        self.dest_shape = (vj.config['rcvr_dim'],\
                            vj.config['pe_dim'],\
                            vj.config['ro_dim'],\
                            vj.config['slc_dim'],\
                            vj.config['et_dim'])
        self.pre_kspace = np.vectorize(complex)(fid_data[:,0::2],\
                                                fid_data[:,1::2])
        self.pre_kspace = np.array(self.pre_kspace,dtype='complex64')
        # check for arrayed parameters, save the length for later 
        self.array_length = vj.util.calc_array_length(fid_data.shape,procpar)
        self.blocks = fid_data.shape[0] // self.array_length
        if verbose:
            print('Making k-space for '+ str(apptype)+' '+str(self.p['seqfil'])+\
                ' seqcon: '+str(self.p['seqcon']))

    def print_fid_header(self):
        for item in self.fhdr.keys():
            print(str('{} : {}').format(item, self.fhdr[item]))

    def make(self):
        """Build k-space from fid data

        Return: 
            kspace=numpy.ndarray([rcvrs,phase,readout,slice,echo/time])
        """
        def _is_interleaved(ppdict):
            res  = (int(ppdict['sliceorder']) == 1)
            return res
        def _is_evenslices(ppdict):
            try:
                res = (int(ppdict['ns']) % 2 == 0)
            except:
                res = (int(ppdict['pss']) % 2 == 0)
            return res
        def make_im2D():
            """Child method of 'make', provides the same as vnmrj im2Drecon"""
    
            p = self.p
            rcvrs = int(p['rcvrs'].count('y'))
            (read, phase, slices) = (int(p['np'])//2, \
                                            int(p['nv']), \
                                            int(p['ns']))
            shiftaxis = (self.config['pe_dim'],self.config['ro_dim'])
            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1

            time = 1
            finalshape = (rcvrs, phase, read, slices,echo*time*self.array_length)
            final_kspace = np.zeros(finalshape,dtype='complex64')
           
            for i in range(self.array_length):
 
                kspace = self.pre_kspace[i*self.blocks:(i+1)*self.blocks,...]

                if p['seqcon'] == 'nccnn':
                    shape = (self.rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], self.dest_shape)

                elif p['seqcon'] == 'nscnn':

                    raise(Exception('not implemented'))

                elif p['seqcon'] == 'ncsnn':

                    preshape = (self.rcvrs, phase, slices*echo*time*read)
                    shape = (self.rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, preshape, order='F')
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], self.dest_shape)

                elif p['seqcon'] == 'ccsnn':

                    preshape = (self.rcvrs, phase, slices*echo*time*read)
                    shape = (self.rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, preshape, order='F')
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], self.dest_shape)
                else:
                    raise(Exception('Not implemented yet'))
                if _is_interleaved(p): # 1 if interleaved slices
                    if _is_evenslices(p):
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:slices//2,:]
                        c[...,1::2,:] = kspace[...,slices//2:,:]
                        kspace = c
                    else:
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:(slices+1)//2,:]
                        c[...,1::2,:] = kspace[...,(slices-1)//2+1:,:]
                        kspace = c

                final_kspace[...,i*echo*time:(i+1)*echo*time] = kspace
            self.kspace = final_kspace
            return final_kspace

        def make_im2Dcs():
            """
            These (*cs) are compressed sensing variants
            """

            def decode_skipint_2D(skipint):

               pass
 
            raise(Exception('not implemented'))

        def make_im2Depi():

            p = self.p
            kspace = self.pre_kspace
            print(kspace.shape)
            nseg = p['nseg']
            kzero = int(p['kzero'])  
            images = int(p['images'])  # repetitions
            time = images
            if p['navigator'] == 'y':
                pluspe = 1 + int(p['nnav'])  # navigator echo + unused
            else:
                pluspe = 1  # unused only
            
            print('images {}'.format(images))
            print('nseg {}'.format(nseg))
            print('ns {}'.format(p['ns']))
            
            if p['pro'] != 0:
                (read, phase, slices) = (int(p['nread']), \
                                            int(p['nphase']), \
                                            int(p['ns']))
            else:
                (read, phase, slices) = (int(p['nread'])//2, \
                                            int(p['nphase']), \
                                            int(p['ns']))
            
            if p['seqcon'] == 'ncnnn':

                preshape = (self.rcvrs, phase+pluspe, slices, time, read)
                print(kspace.size)
                tmp = np.zeros(preshape)
                print(tmp.size)
                kspace = np.reshape(kspace, preshape, order='c')

        def make_im2Depics():
            raise(Exception('not implemented'))
        def make_im2Dfse():
            warnings.warn('May not work correctly')
            kspace = self.pre_kspace
            p = self.p
            petab = vj.util.getpetab(self.procpar,is_procpar=True)
            nseg = int(p['nseg'])  # seqgments
            etl = int(p['etl'])  # echo train length
            kzero = int(p['kzero'])  
            images = int(p['images'])  # repetitions
            (read, phase, slices) = (int(p['np'])//2, \
                                            int(p['nv']), \
                                            int(p['ns']))
            shiftaxis = (self.config['pe_dim'],self.config['ro_dim'])

            echo = 1
            time = images

            phase_sort_order = np.reshape(np.array(petab),petab.size,order='C')
            # shift to positive
            phase_sort_order = phase_sort_order + phase_sort_order.size//2-1

            if p['seqcon'] == 'nccnn':

                #TODO check for images > 1
                preshape = (self.rcvrs, phase//etl, slices, echo*time, etl, read)
                shape = (self.rcvrs, echo*time, slices, phase, read)
                kspace = np.reshape(kspace, preshape, order='C')
                kspace = np.swapaxes(kspace,1,3)
                kspace = np.reshape(kspace, shape, order='C')
                # shape is [rcvrs, phase, slices, echo*time, read]
                kspace = np.swapaxes(kspace,1,3)
                kspace_fin = np.zeros_like(kspace)
                kspace_fin[:,phase_sort_order,:,:,:] = kspace
                kspace_fin = np.moveaxis(kspace_fin, [0,1,4,2,3], self.dest_shape)
                kspace = kspace_fin
            else:
                raise(Exception('not implemented'))
            
            if _is_interleaved(p): # 1 if interleaved slices
                if _is_evenslices(p):
                    c = np.zeros(kspace.shape, dtype='complex64')
                    c[...,0::2,:] = kspace[...,:slices//2,:]
                    c[...,1::2,:] = kspace[...,slices//2:,:]
                    kspace = c
                else:
                    c = np.zeros(kspace.shape, dtype='complex64')
                    c[...,0::2,:] = kspace[...,:(slices+1)//2,:]
                    c[...,1::2,:] = kspace[...,(slices-1)//2+1:,:]
                    kspace = c

            self.kspace = kspace
            return kspace

        def make_im2Dfsecs():
            raise(Exception('not implemented'))
        def make_im3D():

            kspace = self.pre_kspace
            p = self.p
            (read, phase, phase2) = (int(p['np'])//2, \
                                    int(p['nv']), \
                                     int(p['nv2']))

            shiftaxis = (self.config['pe_dim'],\
                        self.config['ro_dim'],\
                        self.config['pe2_dim'])

            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1

            time = 1

            if p['seqcon'] == 'nccsn':
            
                preshape = (self.rcvrs,phase2,phase*echo*time*read)
                shape = (self.rcvrs,phase2,phase,echo*time,read)
                kspace = np.reshape(kspace,preshape,order='F')
                kspace = np.reshape(kspace,shape,order='C')
                kspace = np.moveaxis(kspace, [0,2,4,1,3], self.dest_shape)
                kspace = np.flip(kspace,axis=3)

            if p['seqcon'] == 'ncccn':
                preshape = (self.rcvrs,phase2,phase*echo*time*read)
                shape = (self.rcvrs,phase,phase2,echo*time,read)
                kspace = np.reshape(kspace,preshape,order='F')
                kspace = np.reshape(kspace,shape,order='C')
                kspace = np.moveaxis(kspace, [0,2,4,1,3], self.dest_shape)
    
            if p['seqcon'] == 'cccsn':
            
                preshape = (self.rcvrs,phase2,phase*echo*time*read)
                shape = (self.rcvrs,phase,phase2,echo*time,read)
                kspace = np.reshape(kspace,preshape,order='F')
                kspace = np.reshape(kspace,shape,order='C')
                kspace = np.moveaxis(kspace, [0,2,4,1,3], self.dest_shape)

            if p['seqcon'] == 'ccccn':
                
                shape = (self.rcvrs,phase2,phase,echo*time,read)
                kspace = np.reshape(kspace,shape,order='C')
                kspace = np.moveaxis(kspace, [0,2,4,1,3], self.dest_shape)

            self.kspace = kspace
            return kspace

        def make_im3Dcs():
            """
            3D compressed sensing
            sequences : ge3d, mge3d, se3d, etc
            """
            # -------------------im3Dcs Make helper functions ---------------------

            def decode_skipint_3D(skipint):
                """
                Takes 'skipint' parameter and returns a 0-1 matrix according to it
                which tells what lines are acquired in the phase1-phase2 plane
                """
                BITS = 32  # Skipint parameter is 32 bit encoded binary, see spinsights
                skip_matrix = np.zeros([int(p['nv']), int(p['nv2'])])
                skipint = [int(x) for x in skipint]
                skipint_bin_vals = [str(np.binary_repr(d, BITS)) for d in skipint]
                skipint_bin_vals = ''.join(skipint_bin_vals)
                skipint_bin_array = np.asarray([int(i) for i in skipint_bin_vals])
                skip_matrix = np.reshape(skipint_bin_array, skip_matrix.shape)

                return skip_matrix

            def fill_kspace_3D(pre_kspace, skip_matrix, shape):
                """
                Fills up reduced kspace with zeros according to skip_matrix
                returns zerofilled kspace in the final shape
                """
                kspace = np.zeros(shape, dtype=complex)
                if self.p['seqcon'] == 'ncccn':

                    n = int(self.p['nv'])
                    count = 0
                    for i in range(skip_matrix.shape[0]):
                        for k in range(skip_matrix.shape[1]):
                            if skip_matrix[i,k] == 1:
                                kspace[:,i,k,:,:] = pre_kspace[:,count,:,:]
                                count = count+1
                self.kspace = kspace
                return kspace

            #------------------------im3Dcs make start -------------------------------

            kspace = self.pre_kspace
            p = self.p
            (read, phase, phase2) = (int(p['np'])//2, \
                                    int(p['nv']), \
                                     int(p['nv2']))

            shiftaxis = (self.config['pe_dim'],\
                        self.config['ro_dim'],\
                        self.config['pe2_dim'])

            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1

            time = 1

            if p['seqcon'] == 'nccsn':

                pass

            if p['seqcon'] == 'ncccn':
            
                skip_matrix = decode_skipint_3D(p['skipint'])
                pre_phase = int(self.fid_header['ntraces'])    
                shape = (self.rcvrs, phase, phase2, echo*time, read)
                pre_shape = (self.rcvrs, pre_phase, echo*time, read)
                pre_kspace = np.reshape(kspace, pre_shape, order='c')
                kspace = fill_kspace_3D(pre_kspace, skip_matrix, shape)
                kspace = np.moveaxis(kspace, [0,1,4,2,3],[0,1,2,3,4])
                
            self.kspace = kspace
            return kspace

        def make_im3Dute():
            raise(Exception('not implemented')) 
        # ----------------Handle sequence exceptions first---------------------

        if self.verbose == True:
            print('KspaceMaker: making seqfil : {}'.format(self.p['seqfil']))


        if str(self.p['seqfil']) == 'ge3d_elliptical':
           
            kspace = make_im3Dcs()

        #--------------------------Handle by apptype---------------------------

        if self.p['apptype'] == 'im2D':

            kspace = make_im2D()

        elif self.p['apptype'] == 'im2Dcs':

            kspace = make_im2Dcs()

        elif self.p['apptype'] == 'im2Depi':

            kspace = make_im2Depi()

        elif self.p['apptype'] == 'im2Depics':

            kspace = make_im2Depics()

        elif self.p['apptype'] == 'im2Dfse':

            kspace = make_im2Dfse()

        elif self.p['apptype'] == 'im2Dfsecs':

            kspace = make_im2Dfsecs()

        elif self.p['apptype'] == 'im3D':

            kspace = make_im3D()

        elif self.p['apptype'] == 'im3Dcs':

            kspace = make_im3Dcs()

        elif self.p['apptype'] == 'im3Dute':

            kspace = make_im3Dute()

        else:
            raise(Exception('Could not find apptype. Maybe not implemented?'))

        # ---------------------Global modifications on Kspace------------------

        #TODO if slices are in reversed order, flip them

        return kspace




import vnmrjpy as vj
import numpy as np
import nibabel as nib

class TestFiberFetcher():

    def __init__(self, sequence):

        # sanity check
        if not sequence in ['mems','gems','angio']:
            raise(Exception("incorrenct sequence input. try 'mems', 'angio', etc.."))


        self.sequence = sequence

    def fetch(self):

        def _get_reconpar():
        
            # use aloha init
            pass
    
        if self.sequence == 'angio':
            pass



from numpy import (allclose, angle, arange, argsort, array, asarray,
                   atleast_1d, atleast_2d, cast, dot, exp, expand_dims,
                   iscomplexobj, mean, ndarray, newaxis, ones, pi,
                   poly, polyadd, polyder, polydiv, polymul, polysub, polyval,
                   product, r_, ravel, real_if_close, reshape,
                   roots, sort, take, transpose, unique, where, zeros,
                   zeros_like)
from numpy import (arange, array, asarray, atleast_1d, intc, integer,
                   isscalar, issubdtype, take, unique, where)

import numpy as np


def fftconvolve(in1, in2, mode="full", axes=None):
    """Convolve two N-dimensional arrays using FFT.

    Taken from scipy source
    Convolve `in1` and `in2` using the fast Fourier transform method, with
    the output size determined by the `mode` argument.
    This is generally much faster than `convolve` for large arrays (n > ~500),
    but can be slower when only a few output values are needed, and can only
    output float arrays (int or object array inputs will be cast to float).
    As of v0.19, `convolve` automatically chooses this method or the direct
    method based on an estimation of which is faster.
    Parameters
    ----------
    in1 : array_like
        First input.
    in2 : array_like
        Second input. Should have the same number of dimensions as `in1`.
    mode : str {'full', 'valid', 'same'}, optional
        A string indicating the size of the output:
        ``full``
           The output is the full discrete linear convolution
           of the inputs. (Default)
        ``valid``
           The output consists only of those elements that do not
           rely on the zero-padding. In 'valid' mode, either `in1` or `in2`
           must be at least as large as the other in every dimension.
        ``same``
           The output is the same size as `in1`, centered
           with respect to the 'full' output.
           axis : tuple, optional
    axes : int or array_like of ints or None, optional
        Axes over which to compute the convolution.
        The default is over all axes.
    Returns
    -------
    out : array
        An N-dimensional array containing a subset of the discrete linear
        convolution of `in1` with `in2`.
    """
    in1 = asarray(in1)
    in2 = asarray(in2)
    noaxes = axes is None

    if in1.ndim == in2.ndim == 0:  # scalar inputs
        return in1 * in2
    elif in1.ndim != in2.ndim:
        raise ValueError("in1 and in2 should have the same dimensionality")
    elif in1.size == 0 or in2.size == 0:  # empty arrays
        return array([])

    _, axes = _init_nd_shape_and_axes_sorted(in1, shape=None, axes=axes)

    if not noaxes and not axes.size:
        raise ValueError("when provided, axes cannot be empty")

    if noaxes:
        other_axes = array([], dtype=np.intc)
    else:
        other_axes = np.setdiff1d(np.arange(in1.ndim), axes)

    s1 = array(in1.shape)
    s2 = array(in2.shape)

    if not np.all((s1[other_axes] == s2[other_axes])
                  | (s1[other_axes] == 1) | (s2[other_axes] == 1)):
        raise ValueError("incompatible shapes for in1 and in2:"
                         " {0} and {1}".format(in1.shape, in2.shape))

    complex_result = (np.issubdtype(in1.dtype, np.complexfloating)
                      or np.issubdtype(in2.dtype, np.complexfloating))
    shape = np.maximum(s1, s2)
    shape[axes] = s1[axes] + s2[axes] - 1

    # Check that input sizes are compatible with 'valid' mode
    if _inputs_swap_needed(mode, s1, s2):
        # Convolution is commutative; order doesn't have any effect on output
        in1, s1, in2, s2 = in2, s2, in1, s1

    # Speed up FFT by padding to optimal size for FFTPACK
    fshape = [next_fast_len(d) for d in shape[axes]]
    fslice = tuple([slice(sz) for sz in shape])
    # Pre-1.9 NumPy FFT routines are not threadsafe.  For older NumPys, make
    # sure we only call rfftn/irfftn from one thread at a time.
    if not complex_result and (_rfft_mt_safe or _rfft_lock.acquire(False)):
        try:
            sp1 = np.fft.rfftn(in1, fshape, axes=axes)
            sp2 = np.fft.rfftn(in2, fshape, axes=axes)
            ret = np.fft.irfftn(sp1 * sp2, fshape, axes=axes)[fslice].copy()
        finally:
            if not _rfft_mt_safe:
                _rfft_lock.release()
    else:
        # If we're here, it's either because we need a complex result, or we
        # failed to acquire _rfft_lock (meaning rfftn isn't threadsafe and
        # is already in use by another thread).  In either case, use the
        # (threadsafe but slower) SciPy complex-FFT routines instead.
        sp1 = fftpack.fftn(in1, fshape, axes=axes)
        sp2 = fftpack.fftn(in2, fshape, axes=axes)
        ret = fftpack.ifftn(sp1 * sp2, axes=axes)[fslice].copy()
        if not complex_result:
            ret = ret.real

    if mode == "full":
        return ret
    elif mode == "same":
        return _centered(ret, s1)
    elif mode == "valid":
        shape_valid = shape.copy()
        shape_valid[axes] = s1[axes] - s2[axes] + 1
        return _centered(ret, shape_valid)
    else:
        raise ValueError("acceptable mode flags are 'valid',"
" 'same', or 'full'")

def _init_nd_shape_and_axes(x, shape, axes):
    """Handle shape and axes arguments for n-dimensional transforms.
    Returns the shape and axes in a standard form, taking into account negative
    values and checking for various potential errors.
    Parameters
    ----------
    x : array_like
        The input array.
    shape : int or array_like of ints or None
        The shape of the result.  If both `shape` and `axes` (see below) are
        None, `shape` is ``x.shape``; if `shape` is None but `axes` is
        not None, then `shape` is ``scipy.take(x.shape, axes, axis=0)``.
        If `shape` is -1, the size of the corresponding dimension of `x` is
        used.
    axes : int or array_like of ints or None
        Axes along which the calculation is computed.
        The default is over all axes.
        Negative indices are automatically converted to their positive
        counterpart.
    Returns
    -------
    shape : array
        The shape of the result. It is a 1D integer array.
    axes : array
        The shape of the result. It is a 1D integer array.
    """
    x = asarray(x)
    noshape = shape is None
    noaxes = axes is None

    if noaxes:
        axes = arange(x.ndim, dtype=intc)
    else:
        axes = atleast_1d(axes)

    if axes.size == 0:
        axes = axes.astype(intc)

    if not axes.ndim == 1:
        raise ValueError("when given, axes values must be a scalar or vector")
    if not issubdtype(axes.dtype, integer):
        raise ValueError("when given, axes values must be integers")

    axes = where(axes < 0, axes + x.ndim, axes)

    if axes.size != 0 and (axes.max() >= x.ndim or axes.min() < 0):
        raise ValueError("axes exceeds dimensionality of input")
    if axes.size != 0 and unique(axes).shape != axes.shape:
        raise ValueError("all axes must be unique")

    if not noshape:
        shape = atleast_1d(shape)
    elif isscalar(x):
        shape = array([], dtype=intc)
    elif noaxes:
        shape = array(x.shape, dtype=intc)
    else:
        shape = take(x.shape, axes)

    if shape.size == 0:
        shape = shape.astype(intc)

    if shape.ndim != 1:
        raise ValueError("when given, shape values must be a scalar or vector")
    if not issubdtype(shape.dtype, integer):
        raise ValueError("when given, shape values must be integers")
    if axes.shape != shape.shape:
        raise ValueError("when given, axes and shape arguments"
                         " have to be of the same length")

    shape = where(shape == -1, array(x.shape)[axes], shape)

    if shape.size != 0 and (shape < 1).any():
        raise ValueError(
            "invalid number of data points ({0}) specified".format(shape))

    return shape, axes


def _init_nd_shape_and_axes_sorted(x, shape, axes):
    """Handle and sort shape and axes arguments for n-dimensional transforms.
    This is identical to `_init_nd_shape_and_axes`, except the axes are
    returned in sorted order and the shape is reordered to match.
    Parameters
    ----------
    x : array_like
        The input array.
    shape : int or array_like of ints or None
        The shape of the result.  If both `shape` and `axes` (see below) are
        None, `shape` is ``x.shape``; if `shape` is None but `axes` is
        not None, then `shape` is ``scipy.take(x.shape, axes, axis=0)``.
        If `shape` is -1, the size of the corresponding dimension of `x` is
        used.
    axes : int or array_like of ints or None
        Axes along which the calculation is computed.
        The default is over all axes.
        Negative indices are automatically converted to their positive
        counterpart.
    Returns
    -------
    shape : array
        The shape of the result. It is a 1D integer array.
    axes : array
        The shape of the result. It is a 1D integer array.
    """
    noaxes = axes is None
    shape, axes = _init_nd_shape_and_axes(x, shape, axes)

    if not noaxes:
        shape = shape[axes.argsort()]
        axes.sort()

    return shape, axes


def _inputs_swap_needed(mode, shape1, shape2):
    """
    If in 'valid' mode, returns whether or not the input arrays need to be
    swapped depending on whether `shape1` is at least as large as `shape2` in
    every dimension.
    This is important for some of the correlation and convolution
    implementations in this module, where the larger array input needs to come
    before the smaller array input when operating in this mode.
    Note that if the mode provided is not 'valid', False is immediately
    returned.
    """
    if mode == 'valid':
        ok1, ok2 = True, True

        for d1, d2 in zip(shape1, shape2):
            if not d1 >= d2:
                ok1 = False
            if not d2 >= d1:
                ok2 = False

        if not (ok1 or ok2):
            raise ValueError("For 'valid' mode, one must be at least "
                             "as large as the other in every dimension")

        return not ok1

    return False

def next_fast_len(target):
    """
    Find the next fast size of input data to `fft`, for zero-padding, etc.
    SciPy's FFTPACK has efficient functions for radix {2, 3, 4, 5}, so this
    returns the next composite of the prime factors 2, 3, and 5 which is
    greater than or equal to `target`. (These are also known as 5-smooth
    numbers, regular numbers, or Hamming numbers.)
    Parameters
    ----------
    target : int
        Length to start searching from.  Must be a positive integer.
    Returns
    -------
    out : int
        The first 5-smooth number greater than or equal to `target`.
    Notes
    -----
    .. versionadded:: 0.18.0
    Examples
    --------
    On a particular machine, an FFT of prime length takes 133 ms:
    >>> from scipy import fftpack
    >>> min_len = 10007  # prime length is worst case for speed
    >>> a = np.random.randn(min_len)
    >>> b = fftpack.fft(a)
    Zero-padding to the next 5-smooth length reduces computation time to
    211 us, a speedup of 630 times:
    >>> fftpack.helper.next_fast_len(min_len)
    10125
    >>> b = fftpack.fft(a, 10125)
    Rounding up to the next power of 2 is not optimal, taking 367 us to
    compute, 1.7 times as long as the 5-smooth size:
    >>> b = fftpack.fft(a, 16384)
    """
    hams = (8, 9, 10, 12, 15, 16, 18, 20, 24, 25, 27, 30, 32, 36, 40, 45, 48,
            50, 54, 60, 64, 72, 75, 80, 81, 90, 96, 100, 108, 120, 125, 128,
            135, 144, 150, 160, 162, 180, 192, 200, 216, 225, 240, 243, 250,
            256, 270, 288, 300, 320, 324, 360, 375, 384, 400, 405, 432, 450,
            480, 486, 500, 512, 540, 576, 600, 625, 640, 648, 675, 720, 729,
            750, 768, 800, 810, 864, 900, 960, 972, 1000, 1024, 1080, 1125,
            1152, 1200, 1215, 1250, 1280, 1296, 1350, 1440, 1458, 1500, 1536,
            1600, 1620, 1728, 1800, 1875, 1920, 1944, 2000, 2025, 2048, 2160,
            2187, 2250, 2304, 2400, 2430, 2500, 2560, 2592, 2700, 2880, 2916,
            3000, 3072, 3125, 3200, 3240, 3375, 3456, 3600, 3645, 3750, 3840,
            3888, 4000, 4050, 4096, 4320, 4374, 4500, 4608, 4800, 4860, 5000,
            5120, 5184, 5400, 5625, 5760, 5832, 6000, 6075, 6144, 6250, 6400,
            6480, 6561, 6750, 6912, 7200, 7290, 7500, 7680, 7776, 8000, 8100,
            8192, 8640, 8748, 9000, 9216, 9375, 9600, 9720, 10000)

    target = int(target)

    if target <= 6:
        return target

    # Quickly check if it's already a power of 2
    if not (target & (target-1)):
        return target

    # Get result quickly for small sizes, since FFT itself is similarly fast.
    if target <= hams[-1]:
        return hams[bisect_left(hams, target)]

    match = float('inf')  # Anything found will be smaller
    p5 = 1
    while p5 < target:
        p35 = p5
        while p35 < target:
            # Ceiling integer division, avoiding conversion to float
            # (quotient = ceil(target / p35))
            quotient = -(-target // p35)

            # Quickly find next power of 2 >= quotient
            p2 = 2**((quotient - 1).bit_length())

            N = p2 * p35
            if N == target:
                return N
            elif N < match:
                match = N
            p35 *= 3
            if p35 == target:
                return p35
        if p35 < match:
            match = p35
        p5 *= 5
        if p5 == target:
            return p5
    if p5 < match:
        match = p5
    return match




import unittest
import vnmrjpy as vj
import glob
import nibabel as nib

class Test_SkipintGenerator(unittest.TestCase):

    def test_generate_gems(self):

        reduction = 4
        gemsdir = sorted(glob.glob(vj.fids+'/gems*.fid'))[0]
        procpar = gemsdir+'/procpar'
        gen = vj.util.SkipintGenerator(procpar=procpar) 
        kmask = gen.generate_kspace_mask()
        self.assertEqual(len(kmask.shape),4)
        #nib.viewers.OrthoSlicer3D(kmask).show()

    def test_generate_ge3d(self):

        reduction = 4
        gemsdir = sorted(glob.glob(vj.fids+'/ge3d_s*.fid'))[0]
        procpar = gemsdir+'/procpar'
        gen = vj.util.SkipintGenerator(procpar=procpar) 
        kmask = gen.generate_kspace_mask()
        self.assertEqual(len(kmask.shape),4)
    
        #nib.viewers.OrthoSlicer3D(kmask).show()

    def test_generate_mge3d(self):

        reduction = 4
        gemsdir = sorted(glob.glob(vj.fids+'/mge3d*.fid'))[0]
        procpar = gemsdir+'/procpar'
        gen = vj.util.SkipintGenerator(procpar=procpar) 
        kmask = gen.generate_kspace_mask()
        self.assertEqual(len(kmask.shape),4)
    
        nib.viewers.OrthoSlicer3D(kmask).show()

    def test_skiptab_ge3d(self):
        """ Generate skiptab"""
        pass



import vnmrjpy as vj
import numpy as np
import imageio
import unittest

class Test_Lmafit(unittest.TestCase):

    def test_solve_boat(self):

        img = imageio.imread(vj.pics+'/boat.png')
        mask = np.random.rand(img.shape[0],img.shape[1])
        mask[mask >= 0.7] = 1
        mask[mask < 0.7] = 0
        img_masked = np.multiply(img,mask)

        mse_start = np.abs(((img_masked - img)**2).mean(axis=None))
        self.assertEqual(len(img.shape),2)

        lmafit = vj.aloha.Lmafit(img_masked,\
                    realtimeplot=False,\
                    verbose=False,\
                    tol=5e-3)
        X,Y,out = lmafit.solve()
        img_filled = X.dot(Y)
        mse_end = np.abs(((img_filled - img)**2).mean(axis=None))
        self.assertTrue(mse_end < mse_start)
        print('Starting, ending  MSE : {}, {}'.format(mse_start,mse_end))





"""
Vnmrjpy
=======

"""
import sys
import os
import warnings

# import core functionality

from . import core
from .core import *
# setting global constants not accessible at config
DTYPE='complex64'

# import submodules
from . import aloha
from . import recon
from . import io
from . import util
from . import fit
from . import func
from . import sge

# declare for easy testing

# Read config file
config = util.ConfigParser().parse()




import vnmrjpy as vj
import numpy as np
import nibabel as nib
import warnings
from math import sin, cos

def _set_nifti_header(varr):
    """Set varray.nifti_header attribute for the current space"""
    # init empty header
    header = nib.nifti1.Nifti1Header()
    # use qform only

    # Transform, so q_affine is diagonal
    if varr.space == 'anatomical':

        qfac = 1
        dim_count = len(varr.data.shape)
        header['dim'][0] = dim_count
        for i in range(dim_count):
            header['dim'][i+1] = varr.data.shape[i]
        # setting pixdim based on procpar data
        d = _get_pixdims(varr.pd)
        swap,flip = vj.core.transform._anatomical_swaps(varr.pd)
        d_swap = [i[1] for i in sorted(zip(swap,d))]
        # rotate pixdim
        # see qfac in doc
        header['pixdim'][0] = qfac
        for i in range(3):
            header['pixdim'][i+1] = d_swap[i]

        # set local dim info
        header.set_dim_info(freq=swap[1],phase=swap[0],slice=swap[2])

        # setting qform
        header['qform_code'] = 1
        q_affine = _qform_affine(varr,qfac)
        
        header.set_qform(q_affine,code=1)
   
    # Original without any transformation 
    elif varr.space == 'local':
        
        qfac = 1  # qfac: -1 if left handed
        dim_count = len(varr.data.shape)
        header['dim'][0] = dim_count
        for i in range(dim_count):
            header['dim'][i+1] = varr.data.shape[i]
        # setting pixdom based on procpar data
        d = _get_pixdims(varr.pd)
        
        # see qfac in doc
        header['pixdim'][0] = qfac
        for i in range(3):
            header['pixdim'][i+1] = d[i]

        # set local dim info
        header.set_dim_info(freq=1,phase=0,slice=2)

        # setting qform
        header['qform_code'] = 1
        q_affine = _qform_affine(varr,qfac)
        
        header.set_qform(q_affine,code=1)

    # setting datatype 
    if varr.vdtype == 'complex64': 
        header['datatype'] = 32
    if varr.vdtype == 'float32': 
        header['datatype'] = 16
    if varr.vdtype == 'float64': 
        header['datatype'] = 64

    # units should be mm and sec
    header['xyzt_units'] = 2
    # actually updating header
    varr.nifti_header = header    

    return varr

def _get_pixdims(pd):
    """Return nifti pixdims in [read, phase, slice] dimensions"""

    mul = 10  # procpar lengths are in cm, we need mm
    if pd['apptype'] in ['im2D','im2Dfse','im2Dcs','im2Dfsecs']: 
        d0 = float(pd['lro'])/(int(pd['np'])//2)*mul
        d1 = float(pd['lpe'])/int(pd['nv'])*mul
        d2 = float(pd['thk'])+float(pd['gap'])
        d = (d0,d1,d2)
    elif pd['apptype'] in ['im2Depi','im2Depics']: 
        d0 = float(pd['lro'])/(int(pd['nread'])//2)*mul
        d1 = float(pd['lpe'])/int(pd['nphase'])*mul
        d2 = float(pd['thk'])+float(pd['gap'])
        d = (d0,d1,d2)
    elif pd['apptype'] in ['im3D','im3Dcs','im3Dfse']: 
        d0 = float(pd['lro'])/(int(pd['np'])//2)*mul
        d1 = float(pd['lpe'])/int(pd['nv'])*mul
        d2 = float(pd['lpe2'])/int(pd['nv2'])*mul
        d = (d0,d1,d2)
    else:
        raise(Exception('apptype not implemented in _get_pixdim'))

    return d

def _get_translations(pd):
    """Return translations along axes (read, phase,slice)"""
    # try it naively
    #TODO
    return (0,0,0)

def _qform_affine(varr,qfac):
    """Return qform affine"""
    if varr.space == 'local':
        pixdim = _get_pixdims(varr.pd)
        trans = _get_translations(varr.pd)
        rot = _qform_rot_matrix(varr.pd)
        trans = _translation_to_xyz(trans, rot)
        pixdim_matrix = np.eye(3)*pixdim
        pixdim_matrix[2,2] = pixdim_matrix[2,2]*qfac

        affine = np.zeros((4,4))
        affine[0:3,0:3] = rot @ pixdim_matrix 
        affine[0:3,3] = trans
        affine[3,3] = 1
    elif varr.space == 'anatomical':
        # make the affine diagonal in this case
        warnings.warn('Not in local space, affine probably incorrect')
        pixdim = _get_pixdims(varr.pd)
        trans = _get_translations(varr.pd)
        rot = _qform_rot_matrix(varr.pd)
        pixdim = _pixdim_to_xyz(pixdim, varr.pd)
        # TODO
        #trans = _translation_to_xyz(trans, rot)
        # just make translations 0 for now
        trans = [0,0,0]

        pixdim_matrix = np.eye(3)*pixdim
        pixdim_matrix[2,2] = pixdim_matrix[2,2]*qfac
        aff_3x3 = np.eye(3) * pixdim_matrix

        affine = np.zeros((4,4))
        affine[0:3,0:3] = aff_3x3
        affine[0:3,3] = trans
        affine[3,3] = 1
        #print(affine)
    else:
        raise(Exception('Cant calc q_form affine in this space'))
    return affine


def _qform_rot_matrix(pd):
    """Return rotation matrix for qform affine"""
    # try without
    #TODO
    psi = float(pd['psi']) * 2*np.pi / 360
    phi = float(pd['phi']) * 2*np.pi / 360
    t = float(pd['theta']) * 2*np.pi / 360
    
    #Rotations from Euler angles
    rot_x = np.array([[1,       0,       0],\
                    [0, cos(psi),-sin(psi)],\
                    [0, sin(psi), cos(psi)]])

    rot_y = np.array([[cos(phi),0,sin(phi)],\
                    [0,         1,0      ],\
                    [-sin(phi),0, cos(phi)]])

    rot_z = np.array([[cos(t),-sin(t),0],\
                    [sin(t), cos(t), 0],\
                    [0,     0,      1]])
    m = rot_x @ rot_y @ rot_z 

    #matrix = np.eye(3)
    return m

def _translation_to_xyz(t, rot_matrix):
    """Return translation vector in x,y,z from ro,pe,slc"""
    #TODO
    return t

def _pixdim_to_xyz(pixdim, pd):

    swap, flip = vj.core.transform._anatomical_swaps(pd)
    pixdim = [i[1] for i in sorted(zip(swap,pixdim))]
    return pixdim




import vnmrjpy as vj
import unittest
import numpy as np
import glob
import matplotlib.pyplot as plt

class Test_varray_to_kspace(unittest.TestCase):

    # test form im2D

    def test_gems_to_kspace(self):

        vj.config['verbose'] = False
        gems = glob.glob(vj.config['fids_dir']+'/gems*axial_*')[0]
        varr = vj.read_fid(gems)
        varr.to_kspace()
        self.assertEqual(varr.data.shape,(128,128,20,1,4))
        self.assertEqual(varr.space,'local')
        self.assertEqual(varr.vdtype,'kspace')

    def test_mgems_to_kspace(self):

        vj.config['verbose'] = False
        seq = glob.glob(vj.config['fids_dir']+'/mgems*axial_*')[0]
        varr = vj.read_fid(seq)
        varr.to_kspace()
        self.assertEqual(varr.data.shape,(128,128,20,8,4))
        self.assertEqual(varr.space,'local')
        self.assertEqual(varr.vdtype,'kspace')

    # test for im2Dfse

    def test_fsems_to_kspace(self):

        vj.config['verbose'] = False
        seq = glob.glob(vj.config['fids_dir']+'/fsems*axial_*')[0]
        varr = vj.read_fid(seq)
        varr.to_kspace()
        self.assertEqual(varr.data.shape,(128,128,20,1,4))
        self.assertEqual(varr.space,'local')
        self.assertEqual(varr.vdtype,'kspace')

    # test for im3D

    def test_ge3d_to_kspace(self):

        vj.config['verbose'] = False
        seq = glob.glob(vj.config['fids_dir']+'/ge3d*axial_*')[0]
        varr = vj.read_fid(seq)
        varr.to_kspace()
        self.assertEqual(varr.data.shape,(128,128,128,1,4))
        self.assertEqual(varr.space,'local')
        self.assertEqual(varr.vdtype,'kspace')



import numpy as np
import nibabel as nib
from scipy.signal import gaussian
import vnmrjpy as vj

class SkipintDecoder():
    """Handles decoding of varios skipint formats

    skipint is the parameter responsible for choosing which k-space lines are
    acquired during compressed sensing acquisition

    """

    def __init__(self,procpar_dict):
        """Gather parameters from procpar for decoding
        
        Args:
            procpar_dict (dict) -- procpar dictionary, created by ProcparReader 
        """
        #check skiptab
        if 'skiptab' in procpar_dict.keys() and \
            procpar_dict['skiptab'] == 'y':
            pass
        else:
            raise(Exception("skiptab doesn't exist or set to 'n'."))
        self.skipint = procpar_dict['skipint']
        self.p = procpar_dict
        

    def make_skiptensor(self):
        """Main routine, makes a mask of the acquired lines

        Return:
            skiptensor (np.ndarray) -- multidimensional array, corresponding
                                    to a fiber in kspace. values are 1 if
                                    the kspace line is acquired, 0 if not. 

        """
        p = self.p
        if p['pslabel'] == 'ge3d_elliptical':
            bits = 32  # original encoding, see spinsights
            slice_shape = (int(p['nv']),int(p['nv2']))
            skiptensor = np.zeros(slice_shape)
            skipint = [int(x) for x in self.skipint]
            skipint_bin_vals = [str(np.binary_repr(d,bits)) for d in skipint]
            skipint_bin_vals = ''.join(skipint_bin_vals)
            skipint_bin_array = np.array([int(i) for i in skipint_bin_vals])
            skiptensor = np.reshape(skipint_bin_array, slice_shape,order='f')
            return skiptensor
            

        else:
            raise(Exception('only ge3d_elliptical implemented.'))

class SkipintGenerator():
    """Class to handle skipint parameter generation for compressed sensing

    Currently dynamic indexing is not supported. default kspace mask shape is:

                kspace_mask = np.ndarray([phase, read, slice, time])

    Methods:

        generate_kspace_mask(): 
            Makes k-space mask in the form of numpy array, containing 0s and 1s
        generate_skipint():
            Makes skipint parameter, which is an encoded (compressed) version
            of the mask
    """
    def __init__(self,\
                procpar=None,\
                shape=None,\
                seqdim=None,\
                reduction=4,\
                center=1/16,\
                distribution='gauss',\
                verbose=False):
        """Set up shape of mask

        Args:
            procpar (str): path to procpar file
            shape (tuple): shape of kspace, excluding receivers
            seqdim [2,3]: sequence type, 2dim or 3dim
            reduction (int) : subsampling ration
            center (float) : keep center ratio
            distribution (str) : gauss or random #TODO
            verbose (boolean)

        """
        if procpar != None:

            p = vj.io.ProcparReader(procpar).read()
            conf = vj.config

            if '2D' in p['apptype']:
                dim = 2
            elif '3D' in p['apptype']:
                dim = 3
            # TODO repetitions should handle diffusion and arrayed tr as well
            try:
                reps = int(p['images'])
            except:
                reps = 1
            try:
                echos = int(p['ne'])
            except:
                echos = 1

            shape = [1,1,1,1]
            shape[conf['ro_dim']-1] = int(p['np'])//2
            shape[conf['pe_dim']-1] = int(p['nv'])
            shape[conf['et_dim']-1] = echos * reps
            if dim ==2 :            
                shape[conf['slc_dim']-1] = int(p['ns'])
            elif dim == 3:
                shape[conf['pe2_dim']-1] = int(p['nv2'])

            self.shape = shape
            self.dim = dim
            self.reduction = reduction
            self.center = center
            self.conf = conf
            self.verbose = verbose

    def generate_kspace_mask(self):
        """Makes a kspace mask according to init params

        Returns:

            kspace_mask (np.ndarray) kspace mask of values 0 or 1
        """
        def _make_2d_mask(num):
            """make 2d mask"""
            mask2d = np.zeros((self.shape[0], self.shape[1]),dtype=int)
            weights = list(gaussian(self.shape[0], self.shape[0]/6))
            points = np.zeros(self.shape[0],dtype=int)
            center_ones = [int(points.shape[0]/2*(1-self.center)),\
                            int(points.shape[0]/2*(1+self.center))]
            center_ones_num = center_ones[1] - center_ones[0]
            points[center_ones[0]:center_ones[1]] = 1
            num = num - center_ones_num
            if num < 0:
                raise(Exception('No remaining points to choose: too high\
                                 reduction, or too much points in center'))
        
            indices_available = [i for i, x in enumerate(points) if x == 0]
            weights_available = [weights[i] for i in indices_available]
            weights_available_normalized = [float(i)/sum(weights_available) \
                                            for i in weights_available]
            choices = np.random.choice(indices_available,\
                                        int(num),\
                                        p=weights_available_normalized,\
                                        replace=False)

            for index in choices:
                points[index] = 1

            for i in range(self.shape[1]):

                mask2d[:,i] = np.array(points)

            return mask2d

        def _make_3d_mask(num):

            def gaussian_2d(x, y, stddev_x, stddev_y):

                gx = gaussian(x, stddev_x)
                gy = gaussian(y, stddev_y)
                weights = np.outer(gx,gy.T)
                return weights

            def make_circle_mask(x, y, rx, ry, cx=None, cy=None):

                center_ones = np.zeros([x,y])
                grid = np.ogrid[-x/2 : x/2, -y/2 : y/2]
                mask = grid[0]**2 + grid[1]**2 <= rx*ry
                center_ones[mask] = 1
                return center_ones

            mask3d = np.zeros(self.shape,dtype=int)[...,0]
            points_2d = np.zeros([self.shape[0], self.shape[1]],dtype=int)
            weights_2d = gaussian_2d(self.shape[0], self.shape[1],\
                            self.shape[0]/6, self.shape[1]/6)
            radius_x = self.shape[0]*self.center
            radius_y = self.shape[1]*self.center
            center_ones = make_circle_mask(self.shape[0],
                                            self.shape[1],
                                            radius_x,
                                            radius_y)
            # reshaping weights, points into 1d array
            weights_1d = np.reshape(weights_2d, weights_2d.size)
            points_1d = np.reshape(center_ones, points_2d.size)
            num = num - np.count_nonzero(center_ones == 1)
            if num < 0:
                raise(Exception('No remaining points to choose: too high\
                                 reduction, or too much points in center'))
        
            indices_available = [i for i, x in enumerate(points_1d) if x == 0]
            weights_available = [weights_1d[i] for i in indices_available]
            weights_available_normalized = [float(i)/sum(weights_available) \
                                            for i in weights_available]
            choices = np.random.choice(indices_available,\
                                        int(num),\
                                        p=weights_available_normalized,\
                                        replace=False)

            # what is this?!
            for index in choices:
                points_1d[index] = 1

            points_2d = np.reshape(points_1d, points_2d.shape)

            for i in range(self.shape[2]):

                mask3d[:,:,i] = np.array(points_2d)

            return mask3d

        # --------------------- generate k mask MAIN ------------------------

        kspace_mask = np.zeros(self.shape,dtype=int)
        
        if self.dim == 2:

            num = np.floor(self.shape[0]/self.reduction)

            for t in range(self.shape[self.conf['et_dim']-1]):

                for slc in range(self.shape[self.conf['slc_dim']-1]):

                    kspace_mask[:,:,slc,t] = _make_2d_mask(num)

            if self.verbose:
                print('Mask ready, shape: {}'.format(self.kspace_mask.shape))

        elif self.dim ==3:

            num = np.floor(self.shape[self.conf['pe_dim']-1]*\
                            self.shape[self.conf['pe2_dim']-1]/\
                            self.reduction)

            for t in range(self.shape[self.conf['et_dim']-1]):

                kspace_mask[:,:,:,t] = _make_3d_mask(num)
            #TODO currenty this is a debugging hack. better do it properly 
            kspace_mask = np.swapaxes(kspace_mask,1,2)
        else:
            raise(Exception('wrong dimension maybe'))

        return kspace_mask
    
    def generate_skipint(self,kmask):
        """Makes skipint parameter
        
        Args:
            kmask (np.ndarray) mask, generated by generate_kspace_mask()
        Returns:
            skipint
        """
        pass




import numpy as np
import vnmrjpy as vj
import sys
import copy
import matplotlib.pyplot as plt
import time

def pyramidal_k(kspace_fiber,weights,rp,\
                    realtimeplot=False):
    """ Pyramidal decomposition composit function for single k sparsity case.

    Main Aloha unit.
    One pyramidal decomposition and matrix completion is used at each readout
    point. Thus, the different calls for this function are independent and
    could be passed to the grid or GPU or whatever

    Args:
        kspace_fiber
        weights
        rp
    Return:
        kspace_fiber_complete
    """
    kspace_fiber_complete = copy.deepcopy(kspace_fiber)
    if rp['stages'] == 3:
        lmafit_tolerance = vj.config['lmafit_tol']
    else:
        raise(Exception('no lmafit tolerance at different number of stages\
                        than 3. please specify in config or pyramidal'))
    for s in range(rp['stages']):

        weight = weights[s]
        # init known data for the solvers
        fiber_known = vj.aloha.init_kspace_stage(kspace_fiber,s,rp)
        # saving known center
        fiber_known = vj.aloha.apply_kspace_weights(fiber_known,weight)
        hankel_known = vj.aloha.construct_lvl1_hankel(fiber_known,rp['filter_size'])
        # init data to be completed
        kspace_stage = vj.aloha.init_kspace_stage(kspace_fiber_complete,s,rp)
        kspace_stage = vj.aloha.apply_kspace_weights(kspace_stage,weight)
        hankel = vj.aloha.construct_lvl1_hankel(kspace_stage,rp['filter_size'])
       
        if rp['solver'] == 'svt':
            raise(Exception('not implemented'))
        elif rp['solver'] == 'lmafit':
            lmafit = vj.aloha.Lmafit(hankel,\
                                    known_data=hankel_known,\
                                    verbose=False,\
                                    realtimeplot=realtimeplot,\
                                    tol=lmafit_tolerance[s])
            X,Y,obj = lmafit.solve(max_iter=200)
            hankel = vj.aloha.lowranksolvers.admm(X,\
                                                Y.conj().T,\
                                                fiber_known,\
                                                s,\
                                                rp,\
                                                realtimeplot=realtimeplot,\
                                                max_iter=40)
        fiber = vj.aloha.deconstruct_hankel(hankel,s,rp)
        fiber = vj.aloha.remove_kspace_weights(fiber,weight)
        # replacing center
        kspace_fiber_complete = vj.aloha.finish_kspace_stage(\
                                        fiber,kspace_fiber_complete,0,rp)

    return kspace_fiber_complete

def pyramidal_kxky(kspace_fiber,weights,rp,\
                    realtimeplot=False):
    """ Pyramidal decomposition composit function for kx-ky sparsity case.

    Main Aloha unit.
    One pyramidal decomposition and matrix completion is used at each readout
    point. Thus, the different calls for this function are independent and
    could be passed to the grid or GPU or whatever

    Args:
        kspace_fiber
        weights
        rp
    Return:
        kspace_fiber_complete
    """
    kspace_fiber_complete = copy.deepcopy(kspace_fiber)
    if rp['stages'] == 3:
        lmafit_tolerance = vj.config['lmafit_tol']
    #TODO ugly hack correct this....
    elif rp['stages'] == 1:
        lmafit_tolerance = [vj.config['lmafit_tol'][2]]
    else:
        raise(Exception('no lmafit tolerance at different number of stages\
                        than 3. please specify in config or pyramidal'))
    for s in range(rp['stages']):

        for i in range(2):

            weight = weights[2*s+i]
            # init known data for the solvers
            fiber_known = vj.aloha.init_kspace_stage(kspace_fiber,s,rp)
            # saving known center
            fiber_known = vj.aloha.apply_kspace_weights(fiber_known,weight)
            hankel_known = vj.aloha.construct_hankel(fiber_known,rp)
            # init data to be completed
            kspace_stage = vj.aloha.init_kspace_stage(kspace_fiber_complete,s,rp)
            kspace_stage = vj.aloha.apply_kspace_weights(kspace_stage,weight)
            hankel = vj.aloha.construct_hankel(kspace_stage,rp)
           
            if rp['solver'] == 'svt':
                raise(Exception('not implemented'))
            elif rp['solver'] == 'lmafit':
                lmafit = vj.aloha.Lmafit(hankel,\
                                        known_data=hankel_known,\
                                        verbose=False,\
                                        realtimeplot=realtimeplot,\
                                        tol=lmafit_tolerance[s])
                X,Y,obj = lmafit.solve(max_iter=500)
                print('lmafit ready')
                hankel = vj.aloha.lowranksolvers.admm(X,\
                                                    Y.conj().T,\
                                                    fiber_known,\
                                                    s,\
                                                    rp,\
                                                    realtimeplot=realtimeplot)
            fiber = vj.aloha.deconstruct_hankel(hankel,s,rp)
            fiber = vj.aloha.remove_kspace_weights(fiber,weight)
            # this is for replacing zerofreq lines
            if i == 0:
                weight_x = copy.copy(weight)
                fiber_wx = copy.copy(fiber)
            if i == 1:
                weight_y = weight
                fiber_wy = fiber

        # averaging kx and ky completed fibers
        fiber = vj.aloha.avg_xy_fibers(fiber_wx,fiber_wy,weight_x,weight_y)
        # replacing center
        kspace_fiber_complete = vj.aloha.finish_kspace_stage(\
                                        fiber,kspace_fiber_complete,i,rp)

    return kspace_fiber_complete
            

def pyramidal_kt(kspace_fiber,weights,rp,\
                    realtimeplot=False):
    """ Pyramidal decomposition composit function for k-t sparsity.

    Main Aloha unit.
    One pyramidal decomposition and matrix completion is used at each readout
    point. Thus, the different calls for this function are independent and
    could be passed to the grid or GPU or whatever

    Args:
        kspace_fiber
        weights
        rp
    Return:
        kspace_fiber_complete
    """
    kspace_fiber_complete = copy.deepcopy(kspace_fiber)
    if rp['stages'] == 3:
        lmafit_tolerance = vj.config['lmafit_tol']
    else:
        raise(Exception('no lmafit tolerance at different number of stages\
                        than 3. please specify in config or pyramidal'))
    for s in range(rp['stages']):

        weight = weights[s]
        # init known data for the solvers
        fiber_known = vj.aloha.init_kspace_stage(kspace_fiber,s,rp)
        fiber_known = vj.aloha.apply_kspace_weights(fiber_known,weight)
        hankel_known = vj.aloha.construct_hankel(fiber_known,rp)
        # init data to be completed
        kspace_stage = vj.aloha.init_kspace_stage(kspace_fiber_complete,s,rp)
        kspace_stage = vj.aloha.apply_kspace_weights(kspace_stage,weight)
        hankel = vj.aloha.construct_hankel(kspace_stage,rp)
       
        if rp['solver'] == 'svt':
            raise(Exception('not implemented'))
        elif rp['solver'] == 'lmafit':
            lmafit = vj.aloha.Lmafit(hankel,\
                                    known_data=hankel_known,\
                                    verbose=False,\
                                    realtimeplot=realtimeplot,\
                                    tol=lmafit_tolerance[s])
            X,Y,obj = lmafit.solve(max_iter=500)
            """ old
            admm = vj.aloha.Admm(X,Y.conj().T,fiber_known, s,rp,\
                                    realtimeplot=False)
            hankel = admm.solve()
            """
            hankel = vj.aloha.lowranksolvers.admm(X,\
                                                Y.conj().T,\
                                                fiber_known,\
                                                s,\
                                                rp,\
                                                realtimeplot=realtimeplot)

        fiber = vj.aloha.deconstruct_hankel(hankel,s,rp)
        fiber = vj.aloha.remove_kspace_weights(fiber,weight)
        kspace_fiber_complete = vj.aloha.finish_kspace_stage(\
                                    fiber,kspace_fiber_complete,0,rp)
        
    return kspace_fiber_complete
            

# deprecated-------------------------------------------------------------------
def pyramidal_solve_kt(slice3d,\
                    slice3d_orig,\
                    slice3d_shape,\
                    weights_list,\
                    factors,\
                    rp):
    """Solves a k-t slice: dim0=receivers,dim1=kx,dim2=t
    """ 
    #init
    lmafit_tol_list = vj.config['lmafit_tol']
    solver = rp['solver']
    slice3d_cs = copy.deepcopy(slice3d)
    kspace_complete = copy.deepcopy(slice3d)
    kspace_complete_stage = copy.deepcopy(slice3d)

    for s in range(rp['stages']):
        # init from previous stage

        kspace_init, center = vj.aloha.kspace_pyramidal_init(\
                                            slice3d,s)
        kspace_init_zerofilled, center2 = vj.aloha.kspace_pyramidal_init(\
                                            slice3d_cs,s)
        #kspace_weighing     
        kspace_weighted = vj.aloha.apply_pyramidal_weights_kxt(\
                                            kspace_init,\
                                            weights_list[s],\
                                            rp)
        kspace_zerofilled_weighted = vj.aloha.apply_pyramidal_weights_kxt(\
                                            kspace_init,\
                                            weights_list[s],\
                                            rp)
        #hankel formation
        hankel = vj.aloha.compose_hankel_2d(kspace_weighted,rp)
        hankel_zerofilled = vj.aloha.compose_hankel_2d(\
                                kspace_zerofilled_weighted,rp)
        #low rank matrix completion
        if solver == 'svt':
            svtsolver = SVTSolver(hankel,\
                            tau=None,\
                            delta=None,\
                            epsilon=1e-4,\
                            max_iter=500)
            hankel = vj.aloha.svtsolver.solve()
        elif solver == 'lmafit':
            # initialize with LMaFit
            lmafit = vj.aloha.Lmafit(hankel,\
                            known_data=hankel_zerofilled,\
                            verbose=False,\
                            realtimeplot=False,\
                            tol=lmafit_tol_list[s])
            X,Y,obj = lmafit.solve(max_iter=500)
            admm = vj.aloha.Admm(X,Y.H,\
                        kspace_zerofilled_weighted,\
                        slice3d_shape,\
                        s,\
                        rp,\
                        realtimeplot=False)
            hankel = admm.solve()
            #hankel = U.dot(V.H)
        else:
            raise(Exception('wrong solver'))
        # rearrange original from completed hankel
        kspace_weighted = vj.aloha.decompose_hankel_2d(hankel,\
                            slice3d_shape,s,factors,rp)
        kspace_complete_stage = \
                    vj.aloha.remove_pyramidal_weights_kxt(\
                                        kspace_weighted,\
                                        center,\
                                        weights_list[s])
        kspace_complete = vj.aloha.finalize_pyramidal_stage_kt(\
                                kspace_complete_stage,\
                                kspace_complete,\
                                slice3d, s, rp)    
                           
    kspace_complete = vj.aloha.restore_center(kspace_complete, slice3d)

    return kspace_complete

def pyramidal_solve_kxky(slice3d,\
                    slice3d_orig,\
                    slice3d_shape,\
                    weights_list,\
                    factors,\
                    rp):
    """Solves a k-t slice: dim0=receivers,dim1=kx,dim2=t
    """ 
    #init
    lmafit_tol_list = vj.config['lmafit_tol']
    solver = rp['solver']
    slice3d_cs = copy.deepcopy(slice3d)
    kspace_complete = copy.deepcopy(slice3d)
    kspace_complete_stage = copy.deepcopy(slice3d)
    

    for s in range(rp['stages']):
        weights_ind = 0
        # go through each stage twice
        # weighting is done in one direction at a time
        for _ in range(2):
            # init from previous stage

            kspace_init, center = vj.aloha.kspace_pyramidal_init(\
                                                slice3d,s)
            kspace_init_zerofilled, center2 = vj.aloha.kspace_pyramidal_init(\
                                                slice3d_cs,s)
            #kspace_weighing     
            kspace_weighted = vj.aloha.apply_pyramidal_weights_kxky(\
                                                kspace_init,\
                                                weights_list[s],\
                                                weights_ind,\
                                                rp)
            kspace_zerofilled_weighted = vj.aloha.apply_pyramidal_weights_kxky(\
                                                kspace_init,\
                                                weights_list[s],\
                                                weights_ind,\
                                                rp)
            #hankel formation
            hankel = vj.aloha.compose_hankel_2d(kspace_weighted,rp)
            hankel_zerofilled = vj.aloha.compose_hankel_2d(\
                                    kspace_zerofilled_weighted,rp)
            #low rank matrix completion
            if solver == 'svt':
                svtsolver = SVTSolver(hankel,\
                                tau=None,\
                                delta=None,\
                                epsilon=1e-4,\
                                max_iter=500)
                hankel = vj.aloha.svtsolver.solve()
            elif solver == 'lmafit':
                # initialize with LMaFit
                lmafit = vj.aloha.Lmafit(hankel,\
                                known_data=hankel_zerofilled,\
                                verbose=False,\
                                realtimeplot=False,\
                                tol=lmafit_tol_list[s])
                X,Y,obj = lmafit.solve(max_iter=500)
                admm = vj.aloha.Admm(X,Y.H,\
                            kspace_zerofilled_weighted,\
                            slice3d_shape,\
                            s,\
                            rp,\
                            realtimeplot=False)
                hankel = admm.solve()
                #hankel = U.dot(V.H)
            else:
                raise(Exception('wrong solver'))
            # rearrange original from completed hankel
            kspace_weighted = vj.aloha.decompose_hankel_2d(hankel,\
                                slice3d_shape,s,factors,rp)
            kspace_complete_stage = \
                        vj.aloha.remove_pyramidal_weights_kxky(\
                                            kspace_weighted,\
                                            center,\
                                            weights_list[s],\
                                            weights_ind)
            kspace_complete = vj.aloha.finalize_pyramidal_stage(\
                                    kspace_complete_stage,\
                                    kspace_complete,\
                                    slice3d, s, rp)    
            weights_ind = 1
                           
    kspace_complete = vj.aloha.restore_center(kspace_complete, slice3d)

    return kspace_complete



import unittest
import vnmrjpy as vj
import glob


class Test_FidReader(unittest.TestCase):

    def test_read_gems(self):

        fiddir = sorted(glob.glob(vj.fids+'/gems*'))[0]
        fid = fiddir+'/fid'
        procpar = fiddir+'/procpar'
        reader = vj.io.FidReader(fid)
        data, header = reader.read()
        self.assertEqual(len(data.shape),2)



"""
Collection of higher-level functions acting on varrays.
Mainly for general preprocessing.

Includes:

    COMPOSER -- phase match with short echo time reference scan

"""
#TODO
    #T2fit
    #dixon
    #etc

from .composer import *
from .wasabi import *
from .base import *




import vnmrjpy as vj
import numpy as np
import unittest
import glob
import os
import matplotlib.pyplot as plt


class Test_kmake_epip(unittest.TestCase):

    def test_kmake_from_fid(self):

        #epip_dir = sorted(glob.glob(vj.fids+'/epip_s_2018111301_1shot_lowres_axial_0*'))[0]
        epip_dir = sorted(glob.glob(vj.fids+'/epip_s_2018111301_2shot_axial_0*'))[0]
        fid = epip_dir+'/fid'
        procpar = epip_dir+'/procpar'
        reader = vj.io.FidReader(epip_dir)
        reader.print_header()
        fiddata, fidhdr = reader.read()
        print('epip fid data : {}'.format(fiddata.shape))
        
        kmaker = vj.recon.KspaceMaker(fiddata, fidhdr, procpar, verbose=True) 
        kspace = kmaker.make()
        
        out_dir = vj.config['testresults_dir']+'/kmake/epip'
        outfile_name_k = 'epip_test_kspace'
        #outfile_name_i = 'epip_test_imgspace'
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        outpath_k = out_dir+'/'+outfile_name_k
        #outpath_i = out_dir+'/'+outfile_name_i

        #imaker = vj.recon.ImageSpaceMaker(kspace, procpar)
        #imgspace = imaker.make()
        writer = vj.io.NiftiWriter(np.absolute(kspace[1,...]),procpar)
        writer.write(outpath_k)
        #writer = vj.io.NiftiWriter(np.absolute(imgspace[1,...]),procpar)
        #writer.write(outpath_i)




import vnmrjpy as vj
import numpy as np

"""
Collection of basic functions for reconstruction.

"""

def _ifft(data, dims):
    """Take the inverse fourier transform of kspace data"""
    data = np.fft.fftshift(data,axes=dims) 
    if len(dims) == 2:
        data = np.fft.ifft2(data,axes=dims,norm='ortho')
    elif len(dims) == 3:
        data = np.fft.ifftn(data,axes=dims,norm='ortho')
    data = np.fft.ifftshift(data,axes=dims)
    return data

def _fft(data, dims):
    """Take the inverse fourier transform of imagespace data"""
    pass

def ssos(data,axis=4):
    """Return squared sum of squares combination of receiver data"""
    # if 4 dim, just return the absolute
    if len(data.shape) == 4:
        return np.absolute(data)
    else:
        data = np.sqrt(np.sum(np.absolute(data)**2,axis=axis))
        return np.array(data,dtype='float32')
    



"""
Vnmrjpy Aloha
=============
"""

from .hankelutils import *
from .lowranksolvers import *
from .aloha import *
from .lmafit import *
from .pyramidal import *
from .singularvt import *
from .alternatingdmm import *
from .mathutils import *
from .jitutils import *



import numpy as np
import vnmrjpy as vj
import timeit
import time
import copy

class Aloha():
    """Aloha framework for Compressed Sensing

    ref: Jin et al.: A general framework for compresed sensing and parallel
        MRI using annihilation filter based low-rank matrix completion (2016)
    
    Process outline:

        1. K-space weighing
        2. pyramidal decomposition
        3. Hankel matrix formation
        4. Low-rank matrix completion
        5. K-space unweighing
    """
    def __init__(self, kspace_cs,\
                        procpar,\
                        kspace_orig=None,\
                        reconpar=None):
        """Aloha parameter initialization

        Args:
            procpar : path to procpar file
            kspace_cs : zerofilled cs kspace in numpy array
            reconpar: dictionary, ALOHA recon parameters
                    keys:
                        filter_size
                        rcvrs
                        cs_dim
                        recontype

        TODO : kspace_orig is for test purposes only
        """
        def _get_recontype():
            """get 'recontype', which defines Hankel matrix construction

            Return:
                recontype (str)
            """
            if 'cs' in self.p['apptype']:
                raise(Exception('Cs apptypes not ready yet!'))
            else:
                pass
            # these are for testing. real reconpar will come from apptype
            # TODO
            # or procpar
            if 'angio' in self.p['pslabel']:
                recontype = 'kx-ky_angio'
            elif 'mems' in self.p['pslabel']:
                recontype = 'k-t'
            elif 'ge3d' in self.p['pslabel']:
                recontype = 'kx-ky'
            else:
                raise(Exception('Recon type not implemented yet!'))
            return recontype

        def _get_reconpar(recontype,reconpar,kspace_shape):
            """Make 'rp' recon parameters dictionary

            Reconpar is the main parameter repository of the whole Aloha
            framework. It goes into almost all utility functions during
            recon.
            """

            #TODO maybe make a procpar parameter of this?
            if reconpar == None:

                if recontype == 'k-t':
                    filter_size = (7,5)
                    cs_dim = (vj.config['pe_dim'],vj.config['slc_dim'])
                    ro_dim = vj.config['ro_dim']
                    stages = 3
                    fiber_shape = (kspace_shape[vj.config['rcvr_dim']],\
                                    kspace_shape[vj.config['pe_dim']],\
                                    kspace_shape[vj.config['et_dim']])
                elif recontype in ['kx-ky']:
                    filter_size = (11,11)
                    cs_dim = (vj.config['pe_dim'],vj.config['pe2_dim'])
                    ro_dim = vj.config['ro_dim']
                    stages = 3
                    fiber_shape = (kspace_shape[vj.config['rcvr_dim']],\
                                    kspace_shape[vj.config['pe_dim']],\
                                    kspace_shape[vj.config['pe2_dim']])
                elif recontype in ['kx-ky_angio']:
                    filter_size = (11,11)
                    cs_dim = (vj.config['pe_dim'],vj.config['pe2_dim'])
                    ro_dim = vj.config['ro_dim']
                    stages = 1
                    fiber_shape = (kspace_shape[vj.config['rcvr_dim']],\
                                    kspace_shape[vj.config['pe_dim']],\
                                    kspace_shape[vj.config['pe2_dim']])
                else:
                    raise(Exception('Not implemented'))

                # initialize reconstruction parameters for the whole Aloha
                rp = {'filter_size' : filter_size ,\
                            'cs_dim' : cs_dim ,\
                            'ro_dim' : ro_dim, \
                            'rcvrs' : self.p['rcvrs'].count('y') , \
                            'recontype' : recontype,\
                            'timedim' : vj.config['et_dim'],\
                            'stages' : stages,\
                            'fiber_shape':fiber_shape,\
                            'virtualcoilboost' : vj.config['vcboost'],\
                            'solver' : 'lmafit'}
                return rp
    
            else:
                rp_keys = ['filter_size','cs_dim','ro_dim','rcvrs','recontype',\
                        'timedim','stages','virtualcoilboost','solver']
                # check if dictionary is complete
                for key in rp_keys:
                    if key in reconpar.keys():
                        pass
                    else:
                        raise(Exception('Reconpar key {} not given').format(key))

                return reconpar

        self.p = vj.io.ProcparReader(procpar).read() 
        self.recontype = _get_recontype()
        self.rp = _get_reconpar(self.recontype,reconpar, kspace_cs.shape)
        self.conf = vj.config
        self.kspace_cs = np.array(kspace_cs, dtype='complex64')

    def recon(self):
        """Main reconstruction method for Aloha
        
        Returns:
            kspace_completed (np.ndarray) : full, inferred kspace
        """
        #----------------------------INIT---------------------------------
        def virtualcoilboost_(data):
            """virtual coil data augmentation"""
            #TODO sort by apptype
            if self.p['apptype'] in ['im2D','im2Dfse','im2Depi']:
                conj_trans_data = np.flip(np.flip(data,axis=\
                            self.conf['pe_dim']),axis=self.conf['ro_dim'])

                conj_trans_data = np.conjugate(conj_trans_data)

            elif self.p['apptype'] in ['im3D','im3Dfse']:
                raise(Exception('not implemented'))

            return np.concatenate((data,conj_trans_data),\
                                axis=self.conf['rcvr_dim'])

        if self.rp['virtualcoilboost'] == False:
            kspace_completed = copy.deepcopy(self.kspace_cs)
        elif self.rp['virtualcoilboost'] == True:
            self.kspace_cs = virtualcoilboost_(self.kspace_cs)
            print(self.kspace_cs.shape)
            kspace_completed = copy.deepcopy(self.kspace_cs)

        #------------------------------------------------------------------
        #           2D :    k-t ; kx-ky ; kx-ky_angio
        #------------------------------------------------------------------
        if self.rp['recontype'] in ['k-t','kx-ky','kx-ky_angio']:

            #----------------------MAIN INIT----------------------------    
            if self.rp['recontype'] == 'k-t':

                slice3d_shape = (self.kspace_cs.shape[self.conf['rcvr_dim']],\
                                self.kspace_cs.shape[self.conf['pe_dim']],\
                                self.kspace_cs.shape[self.conf['et_dim']])

                x_len = self.kspace_cs.shape[self.rp['cs_dim'][0]]
                t_len = self.kspace_cs.shape[self.rp['cs_dim'][1]]
                #each element of weight list is an array of weights in stage s
                weights_list = vj.aloha.make_pyramidal_weights_kxt(\
                                        x_len, t_len, self.rp)
                factors = vj.aloha.make_hankel_decompose_factors(\
                                        slice3d_shape, self.rp)
            
            if self.rp['recontype'] in ['kx-ky_angio','kx-ky']:           

                slice3d_shape = (self.kspace_cs.shape[self.conf['rcvr_dim']],\
                                self.kspace_cs.shape[self.conf['pe_dim']],\
                                self.kspace_cs.shape[self.conf['pe2_dim']])

                x_len = self.kspace_cs.shape[self.rp['cs_dim'][0]]
                y_len = self.kspace_cs.shape[self.rp['cs_dim'][1]]
                #each element of weight list is an array of weights in stage s
                if self.rp['recontype'] == 'kx-ky_angio':
                    weights_list = vj.aloha.make_pyramidal_weights_kxkyangio(\
                                        x_len, y_len, self.rp)
                elif self.rp['recontype'] == 'kx-ky':
                    weights_list = vj.aloha.make_pyramidal_weights_kxky(\
                                        x_len, y_len, self.rp)
                else:
                    pass
                factors = vj.aloha.make_hankel_decompose_factors(\
                                        slice3d_shape, self.rp)
            
            #------------------MAIN ITERATION----------------------------    
            for slc in range(self.kspace_cs.shape[3]):

                for x in range(self.kspace_cs.shape[self.rp['cs_dim'][0]]):

                    slice3d = self.kspace_cs[:,:,x,slc,:]
                    slice3d_orig = copy.deepcopy(slice3d)
                    # main call for solvers
                    if self.rp['recontype'] == 'k-t':
                        slice3d_completed = vj.aloha.pyramidal_solve_kt(\
                                                        slice3d,\
                                                        slice3d_orig,\
                                                        slice3d_shape,\
                                                        weights_list,\
                                                        factors,\
                                                        self.rp)
                    elif self.rp['recontype'] in ['kx-ky_angio','kx-ky']:
                        slice3d_completed = vj.aloha.pyramidal_solve_kxky(\
                                                        slice3d,\
                                                        slice3d_orig,\
                                                        slice3d_shape,\
                                                        weights_list,\
                                                        factors,\
                                                        self.rp)
                    else:
                        pass
                    kspace_completed[:,:,x,slc,:] = slice3d_completed
            
                    print('slice {}/{} line {}/{} done.'.format(\
                                slc+1,self.kspace_cs.shape[3],\
                                x+1,self.kspace_cs.shape[2]))

            return kspace_completed




import os
import glob
import numpy as np
import nibabel as nib
import math
import vnmrjpy as vj

class NiftiWriter():
    """Class to write Nifti1 files from procpar and image or kspace data.

    Dimensions, orientations in the input data and procpar must match!
    
    INPUT:
            procpar
            data = numpy.ndarray([phase, readout, slice, time])
            to_scanner (boolean) -- write nifti according to scanner coordinates if True
            niftiheader -- accepts premade header data
    METHODS:
            write(out)

                nifti output dimensions are [phase, readout, slice, time]
                nifti affine is created from the procpar data

    """
    def __init__(self, data, procpar, verbose=False,\
                                    output_space='default',\
                                    input_space='local',\
                                    niftiheader=None):
        """Makes affine, and nifti header"""

        # ----------------------------INIT HELPER FUNCTIONS--------------------
        def _make_scanner_affine():
            """Make appropriate affine for Nifti header"""
            affine = vj.util.make_scanner_affine(self.procpar)
            return affine

        # ----------------------------MAIN INIT--------------------------------
        
        self.procpar = procpar
        self.verbose = verbose
        ppr = vj.io.ProcparReader(procpar)
        self.ppdict = ppr.read()
        self.output_space = output_space
        self.input_space = input_space

        # expand to timedim even if time is only 1
        if len(data.shape) == 3:
            self.data = np.expand_dims(data,axis=-1)
        else:
            self.data = data

        #------------------ making the Nifti affine and header-----------------
        
        # which coordinate system to write?
        if niftiheader==None:

            if output_space=='default':
                #default to config file
                output_space = vj.config['default_space']
            # this is the standard
            if input_space=='local' and output_space=='scanner':
                self.data = vj.util.to_scanner_space(self.data, self.procpar)
                self.affine = vj.util.make_scanner_affine(self.procpar)
            elif input_space=='scanner' and output_space=='scanner':
                self.affine = vj.util.make_scanner_affine(self.procpar)
            elif input_space=='local' and output_space=='rat_anatomical':
                self.data = vj.util.to_scanner_space(self.data, self.procpar)
                self.data = vj.util.scanner_to_rat_anatomical(self.data)
                self.affine = vj.util.make_rat_anatomical_affine(self.procpar)
            elif input_space=='scanner' and output_space=='rat_anatomical':
                self.data = vj.util.scanner_to_rat_anatomical(self.data)
                self.affine = vj.util.make_rat_anatomical_affine(self.procpar)
            elif input_space=='local' and output_space=='local':
                print('Warning! Writenifti : Falling back to cubic affine')
                self.affine = vj.util.make_cubic_affine(self.procpar)
            elif input_space=='rat_anatomical' and output_space=='rat_anatomical':
                self.affine = vj.util.make_rat_anatomical_affine(self.procpar)
            else:
                raise(Exception('Not implemented'))

            self.header = self._make_nifti_header()
        else:
            self.header = niftiheader
            self.affine = niftiheader.affine


    def write(self,out):
        """Saves 4D-7D nifti in .nii.gz format

        data = [dim0,dim1,dim2,timedim,receivers,etc...]

        Arguments:
            out -- save path
        """
        if '.nii' in out:
            out_name = str(out)
        elif '.nii.gz' in out:
            out_name = str(out[:-3])
        else:
            out_name = str(out)+'.nii'
        img = nib.Nifti1Image(self.data, self.affine, self.header)
        nib.save(img,out_name)
        os.system('gzip -f '+str(out_name))
        if self.verbose:
            print('WriteNifti : '+str(out_name)+' saved ... ')
    
    #TODO check if this is even needed
    '''
    def write6d(self,out):
        """Saves data in [x,y,z,t,rvr,real/imag] format to nifti

        Args:
            out -- save path
        """
        if len(self.data.shape) != 6:
            raise(Exception('Data not in 6D format'))
        if '.nii' in out:
            out_name = str(out)
        elif '.nii.gz' in out:
            out_name = str(out[:-3])
        else:
            out_name = str(out)+'.nii'
        img = nib.Nifti1Image(self.data, self.affine, self.header)
        nib.save(img,out_name)
        os.system('gzip -f '+str(out_name))
        if self.verbose:
            print('writeNifti : '+str(out_name)+' saved ... ')
    '''

    def _make_nifti_header(self):
        """Make Nifti header from scratch

        """
        #TODO make some distinction somehow
        if self.output_space == 'rat_anatomical':
            qform_code = 1
        elif self.output_space == 'scanner':
            qform_code = 1
        else:
            qform_code = 0
        p = self.ppdict
        swaparr, flipaxis, sliceaxis = vj.util.get_swap_array(p['orient'])
        header = nib.nifti1.Nifti1Header()
        # TODO leave matrix and use data dims?
        matrix_orig, dim_orig = vj.util.make_local_matrix(self.ppdict)
        matrix = vj.util.swapdim(swaparr,matrix_orig)

        header['xyzt_units'] = 2
        header['dim'][0] = len(self.data.shape)
        header['dim'][1] = matrix[0]
        header['dim'][2] = matrix[1]
        header['dim'][3] = matrix[2]
        header['dim'][4] = self.data.shape[3]
        header['intent_name'] = 'THEINTENT'
        header['aux_file'] = self.procpar
        header.set_qform(self.affine, code=qform_code)

        return header



import numpy as np
import nibabel as nib
import vnmrjpy as vj

class CsTestDataGenerator():
    """Makes .cs bundle for testing
   
    Makes a directory of [seqname].cs, containing:
        imgspace_sum.nii.gz
        kspace_mask.nii.gz
        kspace_imag_ch[n].nii.gz
        kspace_real_ch[n].nii.gz
    """
    def __init__(self,fid,procpar,reduction=4):
        """makes ksapce, imgspace, mask data from fid

        FidReader, KspaceMaker should be available for the sequence
        """    
        p = vj.io.ProcparReader(procpar).read()
        gen = vj.util.SkipintGenerator(procpar=procpar,reduction=reduction)
        self.kspace_mask = gen.generate_kspace_mask()
        fid_data, fid_header = vj.io.FidReader(fid,procpar).read()
        self.kspace = vj.recon.KspaceMaker(fid_data, fid_header, procpar).make()
        self.imgspace = vj.recon.ImageSpaceMaker(self.kspace,procpar).make()
        self.procpar = procpar
        

    def generate(self, savedir=None):
        """saves ksapce, imgspace, mask into savedir"""

        saver = vj.io.SaveRecon(self.procpar,\
                                kspace=self.kspace,\
                                imagespace=self.imgspace)

        masksaver = vj.io.SaveRecon(self.procpar,imagespace=self.kspace_mask)

        if savedir != None:
            saver.save(savedir,savetype='fullkspace') 
            saver.save(savedir,savetype='magn')
            masksaver.save(savedir,savetype='mask')
        else:
            raise(Exception('Please specify output directory'))




import sys
import os
import struct
import numpy as np
from argparse import ArgumentParser
import nibabel as nib

class fdfWriter():
    """
    Class to write data to .fdf files readable by Vnmrj for display
    Intended use: run on custum procesing pipeline and take part in writing
    fdf files from fid data in 'curexp' dir in -study time-
    Intended as an accessory only for this use    

    fdf header is in C syntax.
    header is disgustingly hard coded .... 
    """
    def __init__(self, data, procpar, fid_path=None):
        """
        data = numpy.ndarray(phase, read, slice, echo)
    
        procpar = /path/to/procpar
        """
        self.fid = fid_path
        self.procpar = procpar
        self.p = procparReader(procpar).read()
        self.data = data

    def write(self, outdir):

        """
        FDF header definitions are in Vnmrj User Programming
        Reference manual
        """
        try:
            echoes = int(self.p['ne'])
        except:
            echoes = 1

        images = int(self.data.shape[3])//echoes
        slices = int(self.data.shape[2])
        echoes = int(self.p['ne'])

        def write_fdf(outfile, cropped_data):
            """
            Writes single fdf file.

            outfile = path/to/file

            cropped_data = np.ndarray(phase, read, slice, echo)
            """


            # --------------Setting header --------------------------
            
            # fdf should start with magic number string
            magic_number = '#!/usr/local/fdf/startup'
            
            # setting other header parameters
            if self.p['seqfil'] in ['gems','sems','fsems','mgems','epip']:

                _spatial_rank = '2dfov'
                _rank = 2
                _matrix = (int(self.p['np'])//2, int(self.p['nv']))
                _slices = int(self.p['ns'])
                _ro_size = int(self.p['np'])//2
                _pe_size = int(self.p['nv'])
                _gap = float(self.p['gap'])
                _span = (float(self.p['lro']), float(self.p['lpe']))
                _roi = (float(self.p['lro']),\
                        float(self.p['lpe']),\
                        float(self.p['gap'])+float(self.p['thk']))
    
                _abscissa = '{"cm", "cm"}'

                _ordinate = 'intensity'
                _origin = '{0.0,0.0}'
            elif self.p['seqfil'] in ['ge3d', 'ge3d_elliptical']:

                _spatial_rank = '3dfov'
                _rank = 3
                _matrix = (int(self.p['np'])//2,\
                            int(self.p['nv']),\
                            int(self.p['nv2']))
                _ro_size = int(self.p['np'])//2
                _pe_size = int(self.p['nv'])
                _pe2_size = int(self.p['nv2'])

                _span = (float(self.p['lro']),\
                        float(self.p['lpe']),\
                        float(self.p['lpe2']))

                _roi = (float(self.p['lro']),\
                        float(self.p['lpe']),\
                        float(self.p['lpe2']))
                _imagescale = float(1)
                #TODO
                _gap = float(self.p['gap'])
                _slab_no = 1
                _slabs = 1
                _abscissa = '{"cm", "cm", "cm"}'
                _origin = '{0.0,0.0,0.0}'
                _ordinate = '{ "intensity" }'
            else:
                raise Exception('fdfwriter: sequence not implemented yet')

            _TE = float(self.p['te'])*1000
            _te = float(self.p['te'])
            _TR = float(self.p['tr'])*1000
            _tr = float(self.p['tr'])
            _storage = 'float'
            _bits = 32
            _type = 'absval'
            _bigendian = 0
            _sequence = str(self.p['seqfil'])
            #_studyid = str(self.p['studyid'])
            _studyid = 's_2016052401'  # TODO there is some bug....
            _fid = str(self.fid)

            # dunno if needed...

            _display_order = 0
            _image = 1.0

            # TODO
            _orientation = '{0.0,0.0,1.0,1.0,0.0,0.0,0.0,1.0,0.0}'
            _location = '{0.0,0.0,0.0}'
            _nucleus = '{"H1","C13"}'
            _nucfreq = '{399.459406,100.453412}'

            # ------------------- Making header --------------------

            if self.p['seqfil'] in ['gems','sems','fsems','mgems','epip']:

                header_lines = []
                header_lines.append(magic_number)
                header_lines.append('float\trank = {};'.format(_rank))
                header_lines.append('char\t*storage = "{}";'.format(_storage))
                header_lines.append('char\t*type = "{}";'.format(_type))
                header_lines.append('float\tbits = {};'.format(_bits))
                header_lines.append('float\tmatrix[] = {};'.format(_matrix).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\tspan[] = {};'.format(_span).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\troi[] = {};'.format(_roi).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\tro_size = {};'.format(_ro_size))
                header_lines.append('float\tpe_size = {};'.format(_pe_size))
                header_lines.append('float\tgap = {};'.format(_gap))
                header_lines.append('int\tbigendian = {};'.format(_bigendian))
                header_lines.append('char\t*spatial_rank = "{}";'.format(_spatial_rank))
                header_lines.append('char\t*sequence = "{}";'.format(_sequence))
                header_lines.append('char\t*studyid = "{}";'.format(_studyid))
                header_lines.append('char\t*file = "{}";'.format(_fid))
                # maybe TODO

                header_lines.append('int\tdisplay_order = "{}";'.format(_display_order))
                header_lines.append('float\timage = "{}";'.format(_image))

                #TODO
                header_lines.append('float\torientation[] = {};'.format(_orientation))
                header_lines.append('float\tlocation[] = {};'.format(_location))
                header_lines.append('char\t*nucleus[] = {};'.format(_nucleus))
                header_lines.append('char\t*abscissa[] = {};'.format(_abscissa))
                header_lines.append('float\tnucfreq[] = {};'.format(_nucfreq))
                header_lines.append('float\torigin[] = {};'.format(_origin))
                header_lines.append('float\tnucfreq[] = {};'.format(_nucfreq))

            elif self.p['seqfil'] in ['ge3d','ge3d_elliptical', 'mprage3d']:

                header_lines = []
                header_lines.append(magic_number)
                header_lines.append('float\trank = {};'.format(_rank))
                header_lines.append('char\t*storage = "{}";'.format(_storage))
                header_lines.append('char\t*type = "{}";'.format(_type))
                header_lines.append('float\tbits = {};'.format(_bits))
                header_lines.append('float\tmatrix[] = {};'.format(_matrix).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\tspan[] = {};'.format(_span).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\troi[] = {};'.format(_roi).\
                                    replace('(','{').replace(')','}'))
                header_lines.append('float\tro_size = {};'.format(_ro_size))
                header_lines.append('float\tpe_size = {};'.format(_pe_size))
                header_lines.append('float\tpe2_size = {};'.format(_pe2_size))
                header_lines.append('int\tbigendian = {};'.format(_bigendian))
                header_lines.append('char\t*spatial_rank = "{}";'.format(_spatial_rank))
                header_lines.append('char\t*sequence = "{}";'.format(_sequence))
                header_lines.append('char\t*studyid = "{}";'.format(_studyid))
                header_lines.append('char\t*file = "{}";'.format(_fid))
                header_lines.append('float\tgap = {};'.format(_gap))
                header_lines.append('int\tslabs = {};'.format(_slabs))
                header_lines.append('int\tslab_no = {};'.format(_slab_no))
                # maybe TODO

                header_lines.append('float\timagescale = {};'.format(_imagescale))

                #TODO
                header_lines.append('float\torientation[] = {};'.format(_orientation))
                header_lines.append('float\tlocation[] = {};'.format(_location))
                header_lines.append('float\troi[] = {};'.format(_roi))
                header_lines.append('char\t*nucleus[] = {};'.format(_nucleus))
                header_lines.append('char\t*abscissa[] = {};'.format(_abscissa))
                header_lines.append('float\tnucfreq[] = {};'.format(_nucfreq))
                header_lines.append('float\torigin[] = {};'.format(_origin))
                header_lines.append('char\tordinate[] = {};'.format(_ordinate))

            else:
                raise(Exception('fdfWriter: sequence not implemented yet'))

            # ----------------- Making binary data --------------------
            float_data = np.float32(cropped_data).tobytes()
            print(type(float_data))
            #binary_data = struct.pack('<f',b'\x00'+float_data)
            binary_data =b'\x00'+float_data

            # -------------------Writing single fdf----------------------

            with open(outfile,'w') as openfile:
                header = '\n'.join(header_lines)
                openfile.write(header)           

            #binary_data = np.float32(cropped_data).tobytes()

            with open(outfile,'ab') as openfile:

                #openfile.write(b'\x00')
                openfile.write(binary_data)
 

        # ------------------------- WRITE MAIN --------------------------------
        try: 
            os.mkdir(outdir)
        except:
            print('Warning directory {} exists'.format(outdir))
        os.system('cp '+str(self.procpar)+' '+str(outdir)+'/procpar')

        zn = '000'  # helper string
        
        if self.p['seqfil'] in ['gems','sems','fsems','mgems','epip']:

            for i_slc in range(slices):

                for j_imag in range(images):

                    for k_echo in range(echoes):

                        str_slc = zn[:-len(str(i_slc+1))]+str(i_slc+1)
                        str_imag = zn[:-len(str(j_imag+1))]+str(j_imag+1)
                        str_echo = zn[:-len(str(k_echo+1))]+str(k_echo+1)
                        file_name = 'slice'+str_slc+\
                                    'image'+str_imag+\
                                    'echo'+str_echo+'.fdf'
                        file_path = outdir+'/'+file_name
                        try:
                            open(file_path,'x')
                        except:
                            print('Warning: file {} already exists.\n \
                                    Rewriting...'.format(file_path))
                        write_fdf(file_path,\
                                self.data[:,:,i_slc,j_imag*(k_echo+1)+k_echo])


        elif self.p['seqfil'] in ['ge3d','ge3d_elliptical']:

            print('images: {}, echoes: {}'.format(images, echoes))

            for j_imag in range(images):

                for k_echo in range(echoes):

                    str_imag = zn[:-len(str(j_imag+1))]+str(j_imag+1)
                    str_echo = zn[:-len(str(k_echo+1))]+str(k_echo+1)
                    file_name = 'slab001'+\
                                'image'+str_imag+\
                                'echo'+str_echo+'.fdf'
                    file_path = outdir+'/'+file_name
                    try:
                        open(file_path,'x')
                    except:
                        print('Warning: file {} already exists.\n \
                                Rewriting...'.format(file_path))
                    write_fdf(file_path,\
                                self.data[:,:,:,j_imag*(k_echo+1)+k_echo])
                    print('self data shape: {}'.format(self.data.shape))
                    print('Saving 3D fdf: Done')
        else:
            raise(Exception('writeFdf: sequence not implemented'))

if __name__ == '__main__':

    #parser = ArgumentParser()
    #parser.add_argument('inputfile')

    TESTNIFTI = 'tempdata/imake_test.nii.gz'
    TESTPP = 'tempdata/kmake_test.procpar'

    TESTOUT_DIR = 'tempdata/writefdfdir.img'

    FID_PATH = 'tempdata/sems_20160527_01.fid'

    data = nib.load(TESTNIFTI).get_fdata()

    fdfw = fdfWriter(data, TESTPP, FID_PATH)
    fdfw.write(TESTOUT_DIR)




import vnmrjpy as vj
import numpy as np

def disgustingworkaroundforpossiblebug(p):
    """this is the solution for weird interleaving effects

    p is procpar dictionary
    returns boolean

    use when sliceorder is 1 but is not actually interleaved
    rearrange slices in kmake when false
    """
    def _evenslices(p):
        print('slices {}'.format(p['ns']))
        return (int(p['ns']) % 2 == 0)

    # this means interleaved
    if int(p['sliceorder'])==1:

        # switch for sequnces
        #if p['pslabel'][:4] == 'gems':  
        # switch for seqcon
        if p['seqcon'] == 'nccnn':
            
            # switch for even-odd slices
            if _evenslices(p):

                # switch for orientation
                if p['orient'] in ['sag','sag90','trans90','cor','cor90']:
                    return True

                else:
                    return False

            else:
                # switch for orientation
                if p['orient'] in ['sag','sag90','trans','cor','cor90']:
                    return True
                else:
                    return False
        
        return False
    
    # this means no interleave
    else:
        return False



import copy
import vnmrjpy as vj
import numpy as np
import matplotlib.pyplot as plt

class Lmafit():
    """Low-rank matrix fitting algorithm

    Fills missing matrix elements by low rank approximation

    ref.: paper
    """
    def __init__(self,init_data,\
                known_data='NOT GIVEN',\
                tol=None,\
                k=None,\
                rank_strategy=None,\
                verbose=False,\
                realtimeplot=False):
        """Initialize Lmafit, get defaults from vnmrjpy config

        Args:
            init_data (np.ndarray) : matrix to complete. unkown elements can
                    be approximated beforehand
            known_data (np.ndarray) : matrix of the same shape as init_data 
                    with only the known elements, rest are zero
            tol (float) : tolerance of fitting
            rank_strategy (str) : increase or decrease rank

            verbose
            realtimeplot
        """
        conf = vj.config
        if tol == None:
            tol = conf['lmafit_tol']
        if rank_strategy == None:
            rank_strategy = conf['lmafit_rank_strategy']
        if k == None:
            k = conf['lmafit_start_rank']
        if (type(known_data) == str and known_data == 'NOT GIVEN'):
            if init_data[init_data == 0].size < init_data.size / 10:
                raise(Exception('Known data not given'))
            known_data = copy.deepcopy(init_data)

        self.verbose = verbose
        (m,n) = init_data.shape

        datamask = copy.deepcopy(known_data)
        datamask[datamask != 0+0*1j] = 1
        datanrm = np.linalg.norm(init_data,'fro')
        # init
        #Z = np.matrix(init_data)
        Z = init_data
        #X = np.matrix(np.zeros((m,k),dtype='complex64'))
        #TODO check removed matrix ok
        X = np.zeros((m,k),dtype='complex64')
        Y = np.eye(k,n,dtype='complex64')
        Res = np.multiply(init_data,datamask) - known_data
        res = datanrm
        reschg_tol = 0.5*tol
        # parameters for alf
        alf = 0
        increment = 0.5
        #rank estimation parameters
        itr_rank = 0
        minitr_reduce_rank = 5
        maxitr_reduce_rank = 50
        tau_limit = 10
        rank_incr = 3
        rank_max = 50

        self.realtimeplot = realtimeplot
        if realtimeplot == True:
            self.rtplot = vj.util.RealTimeImshow(np.absolute(init_data))


        self.initpars = (init_data,known_data,\
                        m,n,k,tol,rank_strategy,datanrm,\
                        Z,X,Y,Res,res,reschg_tol,alf,increment,itr_rank,\
                        minitr_reduce_rank,maxitr_reduce_rank,tau_limit,\
                        datamask,rank_incr,rank_max)

    def solve(self,max_iter=100):
        """Main iteration

        Completed matrix: Z = X*Y

        Returns :

            X (np.matrix)
            Y (np.matrix) 
            out_list [] : some list of helper outputs
        """
        def rank_check(R,reschg,tol):
            
            # diag = np.diag(R)
            # d_hat = [diag[i]/diag[i+1] for i in range(len(diag)-1)]
            # tau = (len(diag)-1)*max(d_hat)/(sum(d_hat)-max(d_hat))

            if reschg < 10*tol:
                ind_string = 'increase'
            else:
                ind_string = 'stay'
            return ind_string

        def increase_rank(X,Y,Z,rank_incr,rank_max):
            
            k = X.shape[1]
            k_new = min(k+rank_incr,rank_max)

            m = X.shape[0]
            n = Y.shape[1]
            #X_new = np.matrix(np.zeros((m,k_new),dtype='complex64'))
            #Y_new = np.matrix(np.eye(k_new,n,dtype='complex64'))            
            X_new = np.zeros((m,k_new),dtype='complex64')
            Y_new = np.eye(k_new,n,dtype='complex64')          
            X_new[:,:k] = X
            Y_new[:k,:] = Y
            Z_new = X.dot(Y)
            return X_new, Y_new, Z_new

        # -------------------INIT------------------------

        (data,known_data,m,n,k,tol,rank_strategy,datanrm,\
        Z,X,Y,Res,res,reschg_tol,alf,increment,itr_rank,\
        minitr_reduce_rank,maxitr_reduce_rank,tau_limit,\
                    datamask, rank_incr,rank_max) = self.initpars

        # --------------MAIN ITERATION--------------------
        objv = np.zeros(max_iter)
        RR = np.ones(max_iter)

        for iter_ in range(max_iter):
            itr_rank += 1

            X0 = copy.deepcopy(X)
            Y0 = copy.deepcopy(Y)
            Res0 = copy.deepcopy(Res)
            res0 = copy.deepcopy(res)
            Z0 = copy.deepcopy(Z)
            X = Z.dot(Y.conj().T)
            X, R = np.linalg.qr(X)
            Y = X.conj().T.dot(Z)
            Z = X.dot(Y)

            Res = np.multiply(known_data-Z,datamask)
            res = np.linalg.norm(Res,'fro')
            relres = res / datanrm
            ratio = res / res0
            reschg = np.abs(1-ratio)
            RR[iter_] = ratio
            # adjust alf
            if self.verbose == True:

                print('ratio : {}; rank : {}; reschg : {}, alf : {}'\
                        .format(ratio,X.shape[1],reschg, alf))

            if ratio >= 1.0:
                increment = np.max([0.1*alf,0.1*increment])
                X = copy.deepcopy(X0)
                Y = copy.deepcopy(Y0)
                Res = copy.deepcopy(Res0)
                res = copy.deepcopy(res0)
                relres = res / datanrm
                alf = 0
                Z = copy.deepcopy(Z0)
            elif ratio > 0.7:
                increment = max(increment,0.25*alf)
                alf = alf + increment 
            objv[iter_] = relres
            # check stopping
            if ((reschg < reschg_tol) and ((itr_rank > minitr_reduce_rank) \
                                    or (relres < tol))):
                if self.verbose == True:
                    print('Stopping crit achieved')
                break

            # rank adjustment
            rankadjust = rank_check(R,reschg,tol)
            if rankadjust == 'increase':
                X,Y,Z = increase_rank(X,Y,Z,rank_incr,rank_max)

            Zknown = known_data + alf*Res
            Z = Z - np.multiply(Z,datamask) + Zknown

            if self.realtimeplot == True:
                self.rtplot.update_data(np.absolute(Z))

        obj = objv[:iter_]

        return X, Y, [obj, RR, iter_, relres, reschg] 



import unittest
import numpy as np
import vnmrjpy as vj
from scipy.ndimage.filters import convolve
from scipy.signal import fftconvolve
import matplotlib.pyplot as plt

class Test_fftconvolve(unittest.TestCase):

    def test_fftconvolve(self):

        hankel = np.random.rand(50,20)
        kernel = np.fliplr(np.eye(hankel.shape[1]*2-1))
        #kernel = np.ones((hankel.shape[1],hankel.shape[1]))
        (m,n) = hankel.shape
        div = np.array([ [min(n,j+i+1) for j in range(n)] for i in range(m) ])
        div = np.minimum(div,np.flipud(np.fliplr(div)))
        hankel_new = convolve(hankel,kernel,mode='constant',cval=0)
        hankel_new_fft = fftconvolve(hankel,kernel,mode='same')
        hankel_new_fft_own = vj.aloha.fftconvolve(hankel,kernel,mode='same')
        print('hankel new shape {}'.format(hankel_new.shape))

        hankel_new = hankel_new / div
        hankel_new_fft = hankel_new_fft / div



import numpy as np
import matplotlib.pyplot as plt
import copy
import vnmrjpy as vj
import numba
#import cupy as cp
from scipy.ndimage.filters import convolve
from scipy.signal import fftconvolve
"""
Functions for handling Hankel matrices in various ALOHA implementations

"""
DTYPE = 'complex64'

def avg_xy_fibers(fiber_x,fiber_y,weight_x,weight_y):
    """Weighted average of same fiber completed with different weights.

    In kx-ky completion the weighing is done sequentially in x and y
    directions at a fixed stage

    Args:
        fiber_x,fiber_y (np.ndarray) -- unweighted fibers
        weights_x,weight_y (np.ndarray) -- kspace weights
    Return:
        fiber (np.ndarray) -- averaged fiber
    """
    wx = np.absolute(weight_x)
    wy = np.absolute(weight_y)
    fiber = (wx*fiber_x+wy*fiber_y)/(wx+wy)
    return fiber

def replace_zerofreq_kx(fiber,lines):
    """put back old lines where division by zero occours"""

    ind = fiber.shape[1]//2
    radius = (lines.shape[1]-1)//2
    print('radius {}'.format(radius))
    fiber[:,ind-radius:ind+radius+1,:] = lines
    return fiber

def replace_zerofreq_ky(fiber,lines):

    ind = fiber.shape[2]//2
    radius = (lines.shape[2]-1)//2
    fiber[:,:,ind-radius:ind+radius+1] = lines
    return fiber
    

# this is to be deprecated, but works fine
def construct_hankel_2d(slice3d,rp):
    """Make Hankel matrix from 2d+rcvrs data

    This is the old implementation
    INPUT: slice2d_all_rcvrs : numpy.array[receivers, slice]
    OUTPUT: hankel : numpy.array (m-p)*(n-q) x p*q*rcvrs
    """
    slice3d = np.array(slice3d, dtype='complex64')
    # annihilating filter size
    (p,q) = rp['filter_size']
    # hankel m n
    (m,n) = slice3d.shape[1:3]

    for rcvr in range(slice3d.shape[0]):

        slice2d = slice3d[rcvr,...]

        #make inner hankel list
        for j in range(0,n):
            # iterate in outer hankel elements
            for k in range(0,p):
                # iterate inner hankel columns
                col = np.expand_dims(slice2d[k:m-p+k+1,j],axis=1)
                if k == 0:
                    cols = col
                else:
                    cols = np.concatenate([cols,col],axis=1)
            if j == 0:
                hankel = np.expand_dims(cols,axis=0)
            else:
                hankel = np.concatenate(\
                    [hankel, np.expand_dims(cols,axis=0)], axis=0)

        # make outer hankel
        for i in range(q):
            #col = cp.hstack([hankel[i:n-q+i,:,:]])
            col = np.vstack([hankel[k,:,:] for k in range(i,n-q+i+1)])
            if i == 0:
                cols = col
            else:
                cols = np.concatenate([cols,col], axis=1)
        # concatenating along the receivers
        if rcvr == 0:
            hankel_full = cols
        else:
            hankel_full = np.concatenate([hankel_full, cols], axis=1)

    return hankel_full

def lvl2_hankel_average(hankel_full,filter_shape, fiber_shape):

    (rcvrs,m,n) = fiber_shape
    (p,q) = filter_shape
    inner_shape = (m-p+1,p)
    outer_shape = (n-q+1,q)

    # make convolution kernel
    kernel = np.fliplr(np.eye(outer_shape[0],dtype='complex64'))
    # shape of 1 hnakel without receiver concat
    oldshape = (hankel_full.shape[0],hankel_full.shape[1]//rcvrs)
    newshape = (inner_shape[0],outer_shape[0],inner_shape[1],outer_shape[1])
    for rcvr in range(rcvrs):

        hankel = hankel_full[:,p*q*rcvr:p*q*(rcvr+1)]
        print('hankel shape rcvr {}'.format(hankel.shape))
        hankel = np.reshape(hankel,newshape,order='f')
        hankel = np.swapaxes(hankel,1,2)
        # now axis 0,1 represents the inner blocks, 2,3 is the outer
        for ind in range(n):  # n is also the number of antidiagonals

            # find each i, j pair for ind
            pairs = [[ind-j,j] for j in range(ind+1) if (j < q and ind-j < n-q+1)]

            list_to_avg = [hankel[...,pair[0],pair[1]] for pair in pairs]
            avg = np.zeros_like(hankel[...,0,0],dtype='complex64') 
            for inner_matrix in list_to_avg:
            
                avg += fftconvolve(inner_matrix,kernel,mode='same') 
            # divide by the weights
            avg = avg / len(pairs)

            for pair in pairs:

                hankel[...,pair[0],pair[1]] = avg

        hankel = np.reshape(np.swapaxes(hankel,1,2),oldshape,order='f')
        hankel_full[:,p*q*rcvr:p*q*(rcvr+1)] = hankel

    return hankel_full

def lvl2_hankel_weights(filter_shape,fiber_shape):
    """Make weights for the full 2 level hankel matrix.

    Each element of result gives the multiplicity of an element in the 2-level
    hankel structure

    Args:
        filter_shape (tuple) -- ALOHA annihilating filter shape
        fiber_shape (tuple) -- 3D filter shape (receiver,dim1,dim2)
    Return:
        full_weights
    """
    (rcvrs,m,n) = fiber_shape
    (p,q) = filter_shape
    (x1,y1) = (m-p+1,p)  # inner hankel shape
    (x2,y2) = (n-q+1,q)  # outer block shape, if 1 inner hankel is 1 element
    # base weight of an inner hankel
    weights1 = np.array([[min(y1,i+j+1) for j in range(y1)] for i in range(x1)])
    weights1 = np.minimum(weights1,np.flipud(np.fliplr(weights1)))
    # weights to weight the weights (yeah!)
    weights2 = np.array([[min(y2,i+j+1) for j in range(y2)] for i in range(x2)])
    weights2 = np.minimum(weights2,np.flipud(np.fliplr(weights2)))
    print('weights2 shape {}'.format(weights2.shape))
    full_weights = np.tile(weights1,(x2,y2,1,1))
    # now weight the weights with the weights
    full_weights = full_weights * weights2[...,np.newaxis,np.newaxis]
    full_weights = np.swapaxes(full_weights,1,2)
    full_weights = np.reshape(full_weights,(x1*x2,y1*y2))
    # side-by-side concatenation as per receiverdim
    #full_weights = np.concatenate([full_weights for i in range(rcvrs)],axis=1)
    return full_weights

# this is deprecated and slow
def average_hankel(hankel,stage,rp,level=None):
    """Averages elements in matrix according to the enforced hankel shape

    Args:
        hankel
        stage
        rp
    Returns:
        hankel_avg
    """
    def _get_level(rp):
        """get level count of multilevel hankel"""
        if rp['recontype'] in ['kx-ky','kx-ky_angio','k-t']:
            return 2

    def _calc_fiber_shape(stage,rp_recontype,rp_fiber_shape):
        """Return fiber shape at current stage as tuple"""

        if rp_recontype in ['kx-ky','kx-ky_angio']:
            (rcvrs,x,y) = rp_fiber_shape
            return (rcvrs,x//2**stage,y//2**stage)
        elif rp_recontype == 'k-t':
            (rcvrs,x,t) = rp_fiber_shape
            return (rcvrs,x//2**stage,t)
    
    def _make_kernel(hankel_shape):
        """makes convolution kernel to sum anti-diagonals"""

        return np.fliplr(np.eye(hankel_shape[1]*2-1))

    def _make_weights(hankel_shape):

        (m,n) = hankel_shape
        weights = np.array([[min(n,i+j+1) for j in range(n)] for i in range(m)])
        weights = np.minimum(weights,np.flipud(np.fliplr(weights)))
        return weights

    def _avg_hankel_block(hankel_block,kernel,weights):
    
        hankel_block = convolve(hankel_block,kernel,mode='constant',cval=0)
        return hankel_block / weights

    if level == None:
        level = _get_level(rp)

    if level == 1:
        raise(Exception('not implemented'))
    elif level == 2:

        (rcvrs,m,n) = _calc_fiber_shape(stage,rp['recontype'],rp['fiber_shape'])
        (p,q) = rp['filter_size']
        hankel_block_shape = (m-p+1,p)
        # make filter for inner block anti-diagonal averaging
        kernel = _make_kernel(hankel_block_shape)
        weights = _make_weights(hankel_block_shape)
        #for allhankelblockswithgenerator????
        # reshape hankel to 4d
        shape4d = (hankel.shape[0]//hankel_block_shape[0],hankel_block_shape[0],\
                    hankel.shape[1]//hankel_block_shape[1],hankel_block_shape[1])
        hankel4d = np.reshape(hankel,shape4d)
        for i in range(hankel4d.shape[0]):
            for j in range(hankel4d.shape[2]):

                hankel4d[i,:,j,:] = _avg_hankel_block(hankel4d[i,:,j,:],\
                                                    kernel, weights)

        hankel = np.reshape(hankel4d,hankel.shape)
        return hankel
                
                
    elif level == 3:
        raise('not implemented')

def init_kspace_stage(kspace_fiber,stage,rp):
    """Setup kspace for current pyramidal stage

    At each stage except 0, kspace data is taken from previos completed stage.
    In wavelet sparse images the pyramidal decomposition means taking the
    center part of the kspace, the size corresponds to the current stage.
   
    Args:
        kspace_fiber (np.ndarray) --  full kspace fiber 
        stage (int) -- stage number, first stage is 0
        rp (dict) -- recon parameters
    Return:
        kspace_init (np.ndarray) --  
    """
    if stage == 0:
        return kspace_fiber
         
    if rp['recontype'] in ['kx-ky','kx-ky_angio']: 

        kx_ind = kspace_fiber.shape[1]
        ky_ind = kspace_fiber.shape[2]
        kspace_init = kspace_fiber[:,kx_ind//2-kx_ind//2**(stage+1): \
                                    kx_ind//2+kx_ind//2**(stage+1),\
                                    ky_ind//2-ky_ind//2**(stage+1):\
                                    ky_ind//2+ky_ind//2**(stage+1)] 
    elif rp['recontype'] == 'k-t':
        kx_ind = kspace_fiber.shape[1]
        kspace_init = kspace_fiber[:,kx_ind//2-kx_ind//2**(stage+1): \
                                    kx_ind//2+kx_ind//2**(stage+1),:] 
        
    elif rp['recontype'] == 'k':

        kx_ind = kspace_fiber.shape[1]
        kspace_init = kspace_fiber[:,kx_ind//2-kx_ind//2**(stage+1): \
                                    kx_ind//2+kx_ind//2**(stage+1)] 

    else:
        raise(Exception('not implemented'))

    return kspace_init

def finish_kspace_stage(kspace_stage, kspace_full, stage_iter,rp):
    """Complete pyramidal stage, put inferred data back into original position

    TODO : maybe put known center data back here
    currently 4 center lines ar placed back
    in kxky a rectangle is used instead of circle

    Args:
        kspace_stage
        kspace_full
        stage_iter (0 or 1) -- indicator: kx-ky needs 2 runs for one stage
        rp (dict) -- recon parameters
    Return:
        kspace_full
    """
    
    if rp['recontype'] in ['kx-ky','kx-ky_angio']:
         
        pe_dim = 1  #TODO dimensions are not dynamic currently
        pe2_dim = 2
        # center coordinates for putting bakc original center data
        center_ind_pe = kspace_full.shape[pe_dim]//2-1
        center_ind_pe2 = kspace_full.shape[pe2_dim]//2-1
        # coordinates for stage size
        ind_start = kspace_full.shape[pe_dim]//2 - kspace_stage.shape[pe_dim]//2
        ind_end = kspace_full.shape[pe_dim]//2 + kspace_stage.shape[pe_dim]//2
        ind2_start = kspace_full.shape[pe2_dim]//2 - kspace_stage.shape[pe2_dim]//2
        ind2_end = kspace_full.shape[pe2_dim]//2 + kspace_stage.shape[pe2_dim]//2
        # saving original center
        center = copy.deepcopy(kspace_full[:,center_ind_pe-2:center_ind_pe+2,\
                                center_ind_pe2-2:center_ind_pe2+2])
        # putting back known center
        kspace_full[:,ind_start:ind_end,ind2_start:ind2_end] = kspace_stage

        kspace_full[:,center_ind_pe-2:center_ind_pe+2,\
                        center_ind_pe2-2:center_ind_pe2+2] = center
        return kspace_full

    elif rp['recontype'] == 'k-t':

        pe_dim = 1  #TODO dimensions are not dynamic currently
        center_ind_pe = kspace_full.shape[pe_dim]//2-1

        ind_start = kspace_full.shape[pe_dim]//2 - kspace_stage.shape[pe_dim]//2
        ind_end = kspace_full.shape[pe_dim]//2 + kspace_stage.shape[pe_dim]//2
        #original center
        center = copy.deepcopy(kspace_full[:,center_ind_pe-2:center_ind_pe+2,:])
        kspace_full[:,ind_start:ind_end,:] = kspace_stage
        kspace_full[:,center_ind_pe-2:center_ind_pe+2,:] = center

        return kspace_full
    
    elif rp['recontype'] == 'k':

        pe_dim = 1  #TODO dimensions are not dynamic currently
        center_ind_pe = kspace_full.shape[pe_dim]//2-1

        ind_start = kspace_full.shape[pe_dim]//2 - kspace_stage.shape[pe_dim]//2
        ind_end = kspace_full.shape[pe_dim]//2 + kspace_stage.shape[pe_dim]//2
        #original center
        center = copy.deepcopy(kspace_full[:,center_ind_pe-2:center_ind_pe+2])
        kspace_full[:,ind_start:ind_end] = kspace_stage
        kspace_full[:,center_ind_pe-2:center_ind_pe+2] = center

        return kspace_full
    else:
        raise(Exception('not implemented yet'))

def make_kspace_weights(rp):
    """Create weight data for kspace weighting

    In the cases containing kx-ky, the weights are applied sequentially in
    the pyramidal decomposition at each stage in the 2 dimensions, thus the 
    returned weight list is twice the length.

    Args:
        rp (dictionary) -- recon parameters
    Return:
        weight_list (list of np.ndarray) -- list of kspace weights according
                                            to pyramidal stages
    """
    def _tv_weights(w):

        if w != 0:
            we = 1j*w
        else:
            we = 0.000001
        return we

    def _haar_weights(s,w):
        """Return weights for Haar wavelets at stage s, frequence w"""

        if w!=0:
            we = 1/np.sqrt(2**s)*(-1j*2**s*w)/2*\
                (np.sin(2**s*w/4)/(2**s*w/4))**2*np.exp(-1j*2**s*w/2)
        else:
            we = 0.000001  # just something small not to divide by 0 accidently
        return we

    # init for all cases
    weight_list = []
    if rp['virtualcoilboost'] == True:
        rcvrs_eff = rp['rcvrs']*2
    else:
        rcvrs_eff = rp['rcvrs']
    
    # generate for each different case
    if rp['recontype'] == 'k':

        kx_len = rp['fiber_shape'][1]
        for s in range(rp['stages']):
            w_samples = [2*np.pi/kx_len*k for k in\
                         range(-int(kx_len/2**(s+1)),int(kx_len/2**(s+1)))]
            w_arr = np.array([_haar_weights(s,i) for i in w_samples],\
                                                            dtype='complex64')
            w_arr = w_arr[np.newaxis,:]
            weight_list.append(w_arr)

        return weight_list

    elif rp['recontype'] == 'k-t':

        kx_len = rp['fiber_shape'][1]
        kt_len = rp['fiber_shape'][2]
        for s in range(rp['stages']):
            w_samples = [2*np.pi/kx_len*k for k in\
                         range(-int(kx_len/2**(s+1)),int(kx_len/2**(s+1)))]
            w_arr = np.array([_haar_weights(s,i) for i in w_samples],\
                                                            dtype='complex64')
            w_arr = w_arr[np.newaxis,:,np.newaxis]
            w_arr = np.repeat(w_arr,rcvrs_eff,axis=0)
            w_arr = np.repeat(w_arr,kt_len,axis=2)
            weight_list.append(w_arr)

        return weight_list

    elif rp['recontype'] == 'kx-ky':

        kx_len = rp['fiber_shape'][1]
        ky_len = rp['fiber_shape'][2]
        for s in range(rp['stages']):
            
            # make nd.array for dim x
            w_samples = [2*np.pi/kx_len*k for k in\
                         range(-int(kx_len/2**(s+1)),int(kx_len/2**(s+1)))]
            w_arr = np.array([_haar_weights(s,i) for i in w_samples],\
                                                            dtype='complex64')

            w_arr = w_arr[np.newaxis,:,np.newaxis]
            w_arr = np.repeat(w_arr,rcvrs_eff,axis=0)
            w_arr = np.repeat(w_arr,ky_len//2**s,axis=2)
            weight_list.append(w_arr)
            # make nd.array for dim x
            w_samples = [2*np.pi/ky_len*k for k in\
                         range(-int(ky_len/2**(s+1)),int(ky_len/2**(s+1)))]
            w_arr = np.array([_haar_weights(s,i) for i in w_samples],\
                                                            dtype='complex64')
            w_arr = w_arr[np.newaxis,np.newaxis,:]
            w_arr = np.repeat(w_arr,rcvrs_eff,axis=0)
            w_arr = np.repeat(w_arr,kx_len//2**s,axis=1)
            weight_list.append(w_arr)

        return weight_list

    elif rp['recontype'] == 'kx-ky_angio':
        
        kx_len = rp['fiber_shape'][1]
        ky_len = rp['fiber_shape'][2]
        
        # make nd.array for dim x
        w_samples = [2*np.pi/kx_len*k for k in\
                     range(-int(kx_len/2),int(kx_len/2))]
        w_arr = np.array([_tv_weights(i) for i in w_samples],\
                                                        dtype='complex64')

        w_arr = w_arr[np.newaxis,:,np.newaxis]
        w_arr = np.repeat(w_arr,rcvrs_eff,axis=0)
        w_arr = np.repeat(w_arr,ky_len,axis=2)
        weight_list.append(w_arr)
        # make nd.array for dim x
        w_samples = [2*np.pi/ky_len*k for k in\
                     range(-int(ky_len/2),int(ky_len/2))]
        w_arr = np.array([_tv_weights(i) for i in w_samples],\
                                                        dtype='complex64')
        w_arr = w_arr[np.newaxis,np.newaxis,:]
        w_arr = np.repeat(w_arr,rcvrs_eff,axis=0)
        w_arr = np.repeat(w_arr,kx_len,axis=1)
        weight_list.append(w_arr)

        return weight_list

    elif rp['recontype'] == 'kx-ky-t':
        raise(Exception('Not implemented yet!'))

def apply_kspace_weights(kspace_fiber,weight):
    """Mutiply n-D kspace elements with the approppriate weights

    K-space weighing is a major part in Aloha framework. This function
    determines the proper weighting type, creates and applies the weighting.
    "fiber" is the n-dimensional type of a slice from the k-space.
    Example: in k-t case, the fiber is a 3d cut from kspace along the
    [rcvrs,pe,te] dimensions.

    It!s just a multiplication though, maybe error check...

    Args:
        fiber (np.ndarray) -- input kspace part to weigh
        rp (dictionary) -- recon parameter dictionary
    Return:
        fiber_weighted
    """
    try:
        weighted = np.multiply(kspace_fiber,weight) 
    except:
        raise
    
    return weighted 

def remove_kspace_weights(kspace_fiber,weight):
    """Just divide kspace fiber by the weights"""

    try:
        unweighted = np.divide(kspace_fiber,weight) 
    except:
        raise
    
    return unweighted 

def construct_hankel(nd_data, rp, level=None,order=None):
    """Contruct Multilevel Hankel matrix

    Args:
        nd_data : (np.ndarray) -- input n dimensional data. dim 0 is assumed
                                to be the receiver dimension
        rp (dictionary) -- vnmrjpy aloha reconpar dictionary
        order (list) -- multileveling order. list elements are the dimensions
        level (int) -- number of levels (if None it is assumed from rp)

    Returns:
        hankel (np.ndarray(x,y)) 2D multilevel Hankel matrix
    """
    def _construct_hankel_level(block_vector, filt):
        """Create hanekl from a vector of matrices

        Args:
            block_vector (np.ndarray) --  vector of matrices
            filt (int)  -- filter size on dimension according to Hankel level
        Return:
            hankel
        """
        if len(block_vector.shape) == 1:
            lvl=1
            block_vector = np.expand_dims(block_vector,axis=-1) 
            block_vector = np.expand_dims(block_vector,axis=-1) 
        if len(block_vector.shape) == 3:
            lvl=2
        else:
            raise(Exception('weird shape'))
        shape_x = (block_vector.shape[0]-filt+1)*block_vector.shape[1]
        shape_y = block_vector.shape[2]*filt
        hankel = np.zeros((shape_x,shape_y),dtype='complex64')

        row_num = block_vector.shape[0]-filt+1
        col_num = filt
        row_size = block_vector.shape[1]
        col_size = block_vector.shape[2]
        for col in range(col_num):

            for row in range(row_num):

                hankel[row*row_size : (row+1)*row_size,\
                        col*col_size : (col+1)*col_size] = \
                                            block_vector[col+row,:,:]

        return hankel

    nd_data = np.array(nd_data, dtype='complex64')

    # getting level
    if level == None: 
        level = len(rp['filter_size'])

    if level == 1:

        raise(Exception('not implemented'))

    if level == 2:    
        # annihilating filter size
        p, q = rp['filter_size']
        # hankel m n
        m, n = nd_data.shape[1:]
        #init final shape
        shape_x = (m-p+1)*(n-q+1)
        shape_y = p*q*nd_data.shape[0]  # concatenate along receiver dimension
        shape_x_lvl1 = m-p+1
        shape_y_lvl1 = p
        shape_x_rcvr = shape_x
        shape_y_rcvr = shape_y//nd_data.shape[0]
        
        hankel_rcvr = np.zeros((shape_x_rcvr,shape_y_rcvr),dtype='complex64')
        hankel_lvl1_vec = np.zeros((n,shape_x_lvl1,shape_y_lvl1),dtype='complex64')
        hankel = np.zeros((shape_x,shape_y),dtype='complex64')

        for rcvr in range(nd_data.shape[0]):

            for j in range(nd_data.shape[2]):
               
                #vec = [nd_data[rcvr,i,j] for i in range(nd_data.shape[1])]
                #vec = np.array(vec)
                vec = nd_data[rcvr,:,j]
                hankel_lvl1 = _construct_hankel_level(vec,p)
                hankel_lvl1_vec[j,:,:] = hankel_lvl1
            hankel_rcvr = _construct_hankel_level(hankel_lvl1_vec,q)
            hankel[:,rcvr*shape_y_rcvr:(rcvr+1)*shape_y_rcvr] = hankel_rcvr

        return hankel

    if level == 3:
        raise(Exception('not implemented'))

def deconstruct_hankel(hankel,stage,rp):
    """Make the original ndarray from the multilevel Hankel matrix.

    Used upon completion, and also in an ADMM averaging step
    
    Args:
        hankel (np.ndarray) -- input hankel matrix
        rp (dictionary) -- recon parameters
        stage (int) -- current pyramidal stage
    Return:
        nd_data (np.ndarray)

    """
    def _calc_fiber_shape(stage,rp_recontype,rp_fiber_shape):
        """Return fiber shape at current stage as tuple"""

        if rp_recontype in ['kx-ky','kx-ky_angio']:
            (rcvrs,x,y) = rp_fiber_shape
            return (rcvrs,x//2**stage,y//2**stage)
        elif rp_recontype == 'k-t':
            (rcvrs,x,t) = rp_fiber_shape
            return (rcvrs,x//2**stage,t)
        elif rp_recontype == 'k':
            (rcvrs,x) = rp_fiber_shape
            return (rcvrs,x//2**stage)
        else:
            raise(Exception('not implemented'))
    
    def _block_vector_from_col(col, height):
        """Make vector of blocks from a higher level Hankel column"""

        block_num = col.shape[0]//height
        block_x = height
        block_y = col.shape[1]
        vec = np.zeros((block_num,block_x,block_y),dtype='complex64')
        for num in range(block_num):
            vec[num,:,:] = col[num*block_x:(1+num)*block_x,:]
        return vec

    def _deconstruct_hankel_level(hankel,m,p,factors):
        """ Make an array of Hankel building block

        This is the main local function         
        
        Args:
            hankel -- 2 dim array to decompose
            blockshape -- tuple, shape of inner matrices building up the Hankel
        Return:
            vec -- 3 dim array, dim 0 counts the individual blocks
        """
        inner_x = hankel.shape[0]//(m-p+1)
        inner_y = hankel.shape[1]//p
        block_vector = np.zeros((m,inner_x,inner_y),dtype='complex64')
        for col in range(p):
            to_add = np.zeros((m,inner_x,inner_y),dtype='complex64')
            hankel_col = hankel[:,col*inner_y:(col+1)*inner_y]
            vec = _block_vector_from_col(hankel_col,inner_x)
            to_add[col:col+vec.shape[0],:,:] = vec
            block_vector = block_vector + to_add
        # average the lower level hankel blocks 
        hankels_inner = _average_block_vector(block_vector,factors) 
        return hankels_inner

    def _average_block_vector(vec,factors):
        """Return list of lover level hankels"""

        if len(factors.shape) != len(vec.shape):
            factors = factors[:,np.newaxis,np.newaxis]
        hankels = np.divide(vec,factors)
        return hankels

    def _make_factors(n,q):
        """Make array of Hankel block multiplicities for averaging

        Args:
            n -- kspace length in a dimension
            q -- annihilating filter length in same dimension
        Return:
            factors -- np.array of lenth n
        """
        factors = np.zeros(n) # multiplicity of 1 block
        ramp = [i for i in range(1,q)]
        factors[0:len(ramp)] = ramp
        factors[n-len(ramp):] = ramp[::-1]
        factors[factors == 0] = q
        return factors

    # init
    if rp['recontype'] in ['kx-ky','k-t','kx-ky_angio']:
        level = 2
    elif rp['recontype'] in ['k']:
        level = 1
    # work work 
    if level == 1:

        # subinit level
        (rcvrs,m) = _calc_fiber_shape(stage,rp['recontype'],rp['fiber_shape'])
        p = rp['filter_size']  # annihilating filter
        #number of hankel hankel block columns
        cols_rcvr_lvl2 = hankel.shape[1]//(rcvrs*p) 
        # zeroinit final output
        nd_data = np.zeros((rcvrs,m),dtype='complex64')        

        factors_lvl1 = _make_factors(m,p) # multiplicity of 1 block
        
        # decompose level2
        for rcvr in range(rcvrs):

            hankel_rcvr = hankel[:,rcvr*p:(rcvr+1)*p]
            # decompose level1
            elements = _deconstruct_hankel_level(hankel_rcvr,m,p,factors_lvl1)
            nd_data[rcvr,:] = elements[:,0,0]  # dim 1,2 are single

        #print('nd data shape {}'.format(nd_data.shape))
        return nd_data

    elif level == 2:
        
        # subinit level
        (rcvrs,m,n) = _calc_fiber_shape(stage,rp['recontype'],rp['fiber_shape'])
        (p,q) = rp['filter_size']  # annihilating filter
        #number of hankel hankel block columns
        cols_rcvr_lvl2 = hankel.shape[1]//(rcvrs*p) 
        # zeroinit final output
        nd_data = np.zeros((rcvrs,m,n),dtype='complex64')        

        factors_lvl2 = _make_factors(n,q) # multiplicity of 1 block
        factors_lvl1 = _make_factors(m,p) # multiplicity of 1 block
        
        # decompose level2
        for rcvr in range(rcvrs):

            hankel_rcvr = hankel[:,rcvr*p*q:(rcvr+1)*p*q]
            hankels_lvl1 = _deconstruct_hankel_level(hankel_rcvr,n,q,\
                                                            factors_lvl2)
            # decompose level1
            for num in range(hankels_lvl1.shape[0]):
                hank = hankels_lvl1[num,:,:]
                elements = _deconstruct_hankel_level(hank,m,p,factors_lvl1)
                nd_data[rcvr,:,num] = elements[:,0,0]  # dim 1,2 are single

        return nd_data

    elif level == 3:
        raise(Exception('not implemented'))
    else:
        raise(Exception('not implemented'))




import unittest
import glob
import vnmrjpy as vj

class Test_ProcparReader(unittest.TestCase):

    def test_read_pslabel(self):
        procpar = sorted(glob.glob(vj.fids+'/gems*/procpar'))[0]
        ppr = vj.io.ProcparReader(procpar).read()
        self.assertEqual(ppr['pslabel'],'gems')





import numpy as np
import numba


@numba.njit
def avg_lvl1_hankel(hankel_full,rcvrs):
    """Average matrix, assuming hankel structure
    
    Args:
        hankel (np.ndarray) -- matrix to average
        rcvrs (int) -- receiver channels
    Returns:
        hankel (np.ndarray) -- finished matrix
    """
    a = hankel_full.shape[1]//rcvrs
    for rcvr in range(rcvrs):

        hankel = hankel_full[:,rcvr*a:(rcvr+1)*a]

        num_antidiag = hankel.shape[0]+hankel.shape[1]-1
        # get indice pairs of each antidiagonal
        for ind in range(num_antidiag):
            antidiag_pairs = [[ind-j,j] for j in range(ind+1) \
                            if (j<hankel.shape[1] and ind-j < hankel.shape[0])]
            # avegage antidiagonals
            avg = 0
            for pair in antidiag_pairs:
                avg += hankel[pair[0],pair[1]]
            avg = avg / len(antidiag_pairs)
            # putting back
            for pair in antidiag_pairs:
                hankel[pair[0],pair[1]] = avg

        hankel_full[:,rcvr*a:(rcvr+1)*a]

    return hankel_full

@numba.njit
def avg_lvl2_hankel(hankel_full, block_shape, rcvrs):
    """Averages level 2 block hankel matrix

    Args:
        hankel (2D np.ndarray) -- input matrix to average
        block_shape (tuple) -- inner block shape
        rcvrs (int) -- receiver channels
    Returns:
        hankel (2D np.ndarray) -- averaged matrix

    """

    # -----------------init-----------------
    # outer_shape
    (a,b) = (hankel_full.shape[0]//block_shape[0],\
                hankel_full.shape[1]//block_shape[1]//rcvrs)
    (c,d) = block_shape
    # get indice paris for inner block antidiagonals
    pairs_inner = []
    for ind in range(block_shape[0]+block_shape[1]-1):
        pairs_inner.append([[ind-j,j] for j in range(ind+1) \
                        if (j<d and ind-j < c)])
    pairs_outer = []
    for ind in range(a+b-1):
        pairs_outer.append([[ind-j,j] for j in range(ind+1) \
                        if (j<b and ind-j < a)])
    # -------------actual calc--------------
    # iterate over receiver concatenated blocks
    for rcvr in range(rcvrs):

        hankel = hankel_full[:,rcvr*b*d:(rcvr+1)*b*d]

        # iterate over each block antidiagonal
        for ind in range(a+b-1):
            
            # selecting blocks, make a list
            ij_list = pairs_outer[ind]
            block_list = [hankel[ij[0]*c:(ij[0]+1)*c,ij[1]*d:(ij[1]+1)*d] \
                                                for ij in ij_list]
            # now average the blocks
            temp = block_list[0]
            for k in range(1,len(block_list)):
                temp = np.add(temp,block_list[k])
            avgblock = temp / len(block_list)
            #average antidiag in avgblock
            for antidiag in range(c+d-1):
                avg = 0
                for pair in pairs_inner[antidiag]:
                    avg += avgblock[pair[0],pair[1]]

                avg = avg / len(pairs_inner[antidiag])
                for pair in pairs_inner[antidiag]:
                    avgblock[pair[0],pair[1]] = avg

            # average block is ready, putting bak into lvl2 hankel
            for ij in ij_list:
                hankel[ij[0]*c:(ij[0]+1)*c,ij[1]*d:(ij[1]+1)*d] = avgblock
             
        hankel_full[:,rcvr*b*d:(rcvr+1)*b*d]

    return hankel_full

@numba.njit(parallel=True)
def _hankel_lvl1(block_vector, filt):
    """Create inner (level1) hankel from a vector

    Args:
        block_vector (np.ndarray) --  vector of matrices
        filt (int)  -- filter size on dimension according to Hankel level
    Return:
        hankel
    """
    block_vector = np.expand_dims(block_vector,axis=-1) 
    block_vector = np.expand_dims(block_vector,axis=-1) 
    
    shape_x = (block_vector.shape[0]-filt+1)*block_vector.shape[1]
    shape_y = block_vector.shape[2]*filt
    hankel = np.zeros((shape_x,shape_y),numba.complex64)

    row_num = block_vector.shape[0]-filt+1
    col_num = filt
    row_size = block_vector.shape[1]
    col_size = block_vector.shape[2]
    for col in numba.prange(col_num):

        for row in range(row_num):

            hankel[row*row_size : (row+1)*row_size,\
                    col*col_size : (col+1)*col_size] = \
                                        block_vector[col+row,:,:]

    return hankel

@numba.njit(parallel=True)
def _hankel_lvl2(block_vector, filt):
    """Create level2 hankel from a vector of matrices

    Args:
        block_vector (np.ndarray) --  vector of matrices
        filt (int)  -- filter size on dimension according to Hankel level
    Return:
        hankel
    """
    shape_x = (block_vector.shape[0]-filt+1)*block_vector.shape[1]
    shape_y = block_vector.shape[2]*filt
    hankel = np.zeros((shape_x,shape_y),numba.complex64)

    row_num = block_vector.shape[0]-filt+1
    col_num = filt
    row_size = block_vector.shape[1]
    col_size = block_vector.shape[2]
    for col in numba.prange(col_num):

        for row in range(row_num):

            hankel[row*row_size : (row+1)*row_size,\
                    col*col_size : (col+1)*col_size] = \
                                        block_vector[col+row,:,:]

    return hankel

@numba.njit()
def construct_lvl1_hankel(nd_data, filter_size):
    """Contruct Level1 Hankel matrix

    Args:
        nd_data : (np.ndarray) -- input n dimensional data. dim 0 is assumed
                                to be the receiver dimension
        filter_size (int) -- annihilating filter size

    Returns:
        hankel (np.ndarray(x,y)) 2D multilevel Hankel matrix
    """
    # annihilating filter size
    p = filter_size
    # hankel m n
    m = nd_data.shape[1:][0]
    #init final shape
    shape_x = m-p+1
    shape_y = p*nd_data.shape[0]  # concatenate along receiver dimension
    shape_x_lvl1 = m-p+1
    shape_y_lvl1 = p
    shape_x_rcvr = shape_x
    shape_y_rcvr = shape_y//nd_data.shape[0]
   
    #hankel_rcvr = np.zeros((shape_x_rcvr,shape_y_rcvr),'complex64')
    #hankel = np.zeros((shape_x,shape_y),'complex64')

    hankel_rcvr = np.zeros((shape_x_rcvr,shape_y_rcvr),numba.complex64)
    hankel = np.zeros((shape_x,shape_y),numba.complex64)
    
    for rcvr in range(nd_data.shape[0]):

        vec = nd_data[rcvr,:]
        hankel_rcvr = _hankel_lvl1(vec,p)

        hankel[:,rcvr*shape_y_rcvr:(rcvr+1)*shape_y_rcvr] = hankel_rcvr

    return hankel

@numba.njit()
def construct_lvl2_hankel(nd_data, filter_size):
    """Contruct Multilevel Hankel matrix

    Args:
        nd_data : (np.ndarray) -- input n dimensional data. dim 0 is assumed
                                to be the receiver dimension
        rp (dictionary) -- vnmrjpy aloha reconpar dictionary

    Returns:
        hankel (np.ndarray(x,y)) 2D multilevel Hankel matrix
    """
    # annihilating filter size
    p, q = filter_size
    # hankel m n
    m, n = nd_data.shape[1:]
    #init final shape
    shape_x = (m-p+1)*(n-q+1)
    shape_y = p*q*nd_data.shape[0]  # concatenate along receiver dimension
    shape_x_lvl1 = m-p+1
    shape_y_lvl1 = p
    shape_x_rcvr = shape_x
    shape_y_rcvr = shape_y//nd_data.shape[0]
   
    hankel_rcvr = np.zeros((shape_x_rcvr,shape_y_rcvr),numba.complex64)
    hankel_lvl1_vec = np.zeros((n,shape_x_lvl1,shape_y_lvl1),numba.complex64)
    hankel = np.zeros((shape_x,shape_y),numba.complex64)
    
    for rcvr in range(nd_data.shape[0]):

        for j in range(nd_data.shape[2]):
           
            #vec = [nd_data[rcvr,i,j] for i in range(nd_data.shape[1])]
            #vec = np.array(vec)
            vec = nd_data[rcvr,:,j]
            hankel_lvl1 = _hankel_lvl1(vec,p)
            hankel_lvl1_vec[j,:,:] = hankel_lvl1
        hankel_rcvr = _hankel_lvl2(hankel_lvl1_vec,q)
        hankel[:,rcvr*shape_y_rcvr:(rcvr+1)*shape_y_rcvr] = hankel_rcvr

    return hankel
#---------------------------------SLOOWWWWWWWW----------------------------
@numba.njit()
def deconstruct_lvl2_hankel(hankel,stage,recontype_int,fiber_shape,filter_size):
    """Make the original ndarray from the multilevel Hankel matrix.

    Used upon completion, and also in an ADMM averaging step
    
    Args:
        hankel (np.ndarray) -- input hankel matrix
        rp (dictionary) -- recon parameters
        stage (int) -- current pyramidal stage
    Return:
        nd_data (np.ndarray)

    """
    
    if recontype_int == 1:
        (rcvrs,x,y) = fiber_shape
        (rcvrs,m,n) = (rcvrs,x//2**stage,y//2**stage)
    elif recontype_int == 2:
        (rcvrs,x,y) = fiber_shape
        (rcvrs,m,n) = (rcvrs,x//2**stage,y)
    (p,q) = filter_size  # annihilating filter
    #number of hankel hankel block columns
    cols_rcvr_lvl2 = hankel.shape[1]//(rcvrs*p) 
    # zeroinit final output
    nd_data = np.zeros((rcvrs,m,n),numba.complex64)        

    factors_lvl2 = _make_factors(n,q) # multiplicity of 1 block
    factors_lvl1 = _make_factors(m,p) # multiplicity of 1 block
    
    # decompose level2
    for rcvr in range(rcvrs):

        hankel_rcvr = hankel[:,rcvr*p*q:(rcvr+1)*p*q]
        hankels_lvl1 = _deconstruct_hankel_lvl2(hankel_rcvr,n,q,factors_lvl2)
        # decompose level1
        for num in range(hankels_lvl1.shape[0]):
            hank = hankels_lvl1[num,:,:]
            elements = _deconstruct_hankel_lvl1(hank,m,p,factors_lvl1)
            nd_data[rcvr,:,num] = elements[:,0,0]  # dim 1,2 are single

    return nd_data

@numba.njit()
def _calc_fiber_shape(stage,recontype_int,fiber_shape):
    """Return fiber shape at current stage as tuple"""

    if recontype_int == 1:
        (rcvrs,x,y) = fiber_shape
        return (rcvrs,x//2**stage,y//2**stage)
    elif recontype_int == 2:
        (rcvrs,x,t) = fiber_shape
        return (rcvrs,x//2**stage,t)

@numba.njit()
def _block_vector_from_col(col, height):
    """Make vector of blocks from a higher level Hankel column"""

    block_num = col.shape[0]//height
    block_x = height
    block_y = col.shape[1]
    vec = np.zeros((block_num,block_x,block_y),numba.complex64)
    for num in range(block_num):
        vec[num,:,:] = col[num*block_x:(1+num)*block_x,:]
    return vec

@numba.njit()
def _deconstruct_hankel_lvl1(hankel, m, p, factors):
    """ Make an array of Hankel building block

    This is the main local function         
    
    Args:
        hankel -- 2 dim array to decompose
        blockshape -- tuple, shape of inner matrices building up the Hankel
        level (int) -- specifies hankel level 
    Return:
        vec -- 3 dim array, dim 0 counts the individual blocks
    """
    inner_x = hankel.shape[0]//(m-p+1)
    inner_y = hankel.shape[1]//p
    block_vector = np.zeros((m,inner_x,inner_y),numba.complex64)
    for col in range(p):
        to_add = np.zeros((m,inner_x,inner_y),numba.complex64)
        hankel_col = hankel[:,col*inner_y:(col+1)*inner_y]
        vec = _block_vector_from_col(hankel_col,inner_x)
        to_add[col:col+vec.shape[0],:,:] = vec
        block_vector = block_vector + to_add
        # average the lower level hankel blocks
        hankels_inner = np.divide(block_vector,factors)
    return hankels_inner

@numba.njit()
def _deconstruct_hankel_lvl2(hankel, m, p, factors):
    """ Make an array of Hankel building block

    This is the main local function         
    
    Args:
        hankel -- 2 dim array to decompose
        blockshape -- tuple, shape of inner matrices building up the Hankel
        level (int) -- specifies hankel level 
    Return:
        vec -- 3 dim array, dim 0 counts the individual blocks
    """
    inner_x = hankel.shape[0]//(m-p+1)
    inner_y = hankel.shape[1]//p
    block_vector = np.zeros((m,inner_x,inner_y),numba.complex64)
    for col in range(p):
        to_add = np.zeros((m,inner_x,inner_y),numba.complex64)
        hankel_col = hankel[:,col*inner_y:(col+1)*inner_y]
        vec = _block_vector_from_col(hankel_col,inner_x)
        to_add[col:col+vec.shape[0],:,:] = vec
        block_vector = block_vector + to_add
        # average the lower level hankel blocks
        _factors = np.expand_dims(np.expand_dims(factors,axis=-1),axis=-1)
        hankels_inner = np.divide(block_vector,_factors)
    return hankels_inner

@numba.njit()
def _make_factors(n,q):
    """Make array of Hankel block multiplicities for averaging

    Args:
        n -- kspace length in a dimension
        q -- annihilating filter length in same dimension
    Return:
        factors -- np.array of lenth n
    """
    factors = np.zeros(n) # multiplicity of 1 block
    ramp = [i for i in range(1,q)]
    factors[0:len(ramp)] = ramp
    factors[n-len(ramp):] = ramp[::-1]
    factors[factors == 0] = q
    return factors




import vnmrjpy as vj
import numpy as np
import matplotlib.pyplot as plt
import unittest
import nibabel as nib
import glob

class Test_transform_to_scanner(unittest.TestCase):

    """
    def test_transform_gems(self):

        axial = glob.glob(vj.fids+'/gems*axial_0_0_0_01.fid')[0]
        axial90 = glob.glob(vj.fids+'/gems*axial90_0_90_0_01.fid')[0]
        coronal = glob.glob(vj.fids+'/gems*coronal_0_0_90_01.fid')[0]
        gems_list = [axial, axial90, coronal]
        gems_procpars = [i+'/procpar' for i in gems_list]
        gems_names = ['axial', 'axial90', 'coronal']
        gems_data_list = []
        for num, item in enumerate(gems_list):
            print(item)
            itemdata, affine = vj.io.FidReader(item).make_image()
            #itemdata = vj.util.to_scanner_space(itemdata,\
            #                                    gems_procpars[num])
            v = nib.viewers.OrthoSlicer3D(itemdata,affine=affine)
            v.show()
            gems_data_list.append(itemdata)

        print(vj.util.check_90deg(gems_procpars[1]))
    """
    def test_transform_mems(self):

        axial = glob.glob(vj.fids+'/mems*axial_0_0_0_01.fid')[0]
        axial90 = glob.glob(vj.fids+'/mems*axial90_0_90_0_01.fid')[0]
        coronal = glob.glob(vj.fids+'/mems*coronal_0_0_90_01.fid')[0]
        mems_list = [axial, axial90, coronal]
        mems_procpars = [i+'/procpar' for i in mems_list]
        mems_names = ['axial', 'axial90', 'coronal']
        mems_data_list = []
        for num, item in enumerate(mems_list):
            print(item)
            itemdata, affine = vj.io.FidReader(item).make_image()
            #itemdata = vj.util.to_scanner_space(itemdata,\
            #                                    mems_procpars[num])
            v = nib.viewers.OrthoSlicer3D(itemdata,affine=affine)
            v.show()
            mems_data_list.append(itemdata)

        print(vj.util.check_90deg(mems_procpars[1]))
        



import unittest
import vnmrjpy as vj
import glob
import numpy as np
import matplotlib.pyplot as plt

class Test_FdfReader(unittest.TestCase):

    def test_read_img(self):

        fdfdir = sorted(glob.glob(vj.fdfs+'/gems*'))[0]
        
        reader = vj.io.FdfReader(fdfdir)
        data, header = reader.read()
        self.assertEqual(type(data),np.ndarray)
        self.assertEqual(type(header),dict)
        self.assertEqual(len(data.shape),4)
